#include "ransac.h"


void ransac (   targets_data_t tdat, 
                int min_data, 
                int max_iterations,
                double threshold, 
                int close_data_values,
		FILE *ferased)
{

    #define HORIZONTAL 0 
    #define VERTICAL   1 

    int iterations = 0;
    int ndata = tdat->ntargets;
    //double hmin = (double)min_data/2.0;
    //double hdata = (double)tdat->ntargets/2.0;

    int max_npoints_hor = 4;
    int max_npoints_ver = 4;

    targets_data_t inliers_set   = tf_targets_data_create(min_data);
    targets_data_t consensus_set = tf_targets_data_create(ndata);
    targets_data_t best_set      = tf_targets_data_create(ndata);

    double best_error = HUGE;
    double this_error;

    /*Choosing the starting seed to generate de random numbers*/
    srand(time(NULL));

    int tsaiWeights = 0;

    int type[24] = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1 };
    //int type[8] = { 0, 0, 0, 0, 1, 1, 1, 1};
    //int type[18] = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1};

    int d_best[ndata];
    int d_consensus[ndata];

    int found_best = 0;

    struct _calibration_data_t cdat = (struct _calibration_data_t){ tdat->ntargets, tdat->p_w, tdat->p_i };

    //fprintf(stderr, "inliers = %d,  consensus_set = %d, close_data_values = %d\n", inliers_set->ntargets, consensus_set->ntargets, close_data_values);
    while (iterations < max_iterations)
    {

         int k = 0, l;

         int d_inliers[min_data];

	 int npoints_hor = 0, npoints_ver = 0;

         camera_specs_t cspec = tf_get_canon_optura_specs (); 
         camera_parameters_t inliers_cpar = tf_get_mean_camera_parameters(cspec);

         while (k < min_data) {

               int dselected = ( rand() % ndata );
	       if (!is_inside (dselected, min_data, d_inliers)) {
                    /*Choosing data points from vertical*/
	            if ((type[dselected] == HORIZONTAL) && (npoints_hor < max_npoints_hor) && (npoints_hor+npoints_ver) <= min_data) {
	                d_inliers[k] = dselected;
	                npoints_hor++;
			k++;
                    }	   
                    /*Choosing data points from horizontal*/
		    else if ((type[dselected] == VERTICAL) && (npoints_ver < max_npoints_ver) && (npoints_hor+npoints_ver) <= min_data) {
		        d_inliers[k] = dselected;
			k++;
		        npoints_ver++;
                    }
	       }
         }

         /*Coping the selected points to the inliers set*/

         for (k = 0; k < min_data; k++) {
              inliers_set->p_i[k]  = tdat->p_i[d_inliers[k]];
              inliers_set->p_w[k]  = tdat->p_w[d_inliers[k]];
              inliers_set->conf[k] = tdat->conf[d_inliers[k]];      
              d_consensus[k] = d_inliers[k];
         }
         //fprintf(ferased, "\n");


         /*Creating the model*/
	 struct _calibration_data_t inliers_cdat = (struct _calibration_data_t)
                                                   {
                                                         inliers_set->ntargets, 
                                                         inliers_set->p_w, 
                                                         inliers_set->p_i, 
                                                         inliers_set->conf 
                                                   };
         
        //if (! tsaiWeights) {
        //  tf_calib_set_weights(&inliers_cdat, inliers_set->conf);
        //} else {
           tf_calib_set_weights_uniform(&inliers_cdat);
        //}
        tf_calib2_tsai (&inliers_cdat, cspec, inliers_cpar, FALSE, TRUE);

        // tf_write_mutable_camera_parameters (ferased, iterations, inliers_cpar);

         /*Setting consensus set as maybe_inliers*/
         tf_targets_data_copy(inliers_set, consensus_set);

         /*Test each point that do not belongs to the inliers set, and verify 
           if its ajusts well in the camera model found  with the inliers set*/
         int t, inside;
         l = min_data;
	 consensus_set->ntargets = inliers_set->ntargets;
	 for (k = 0; k < ndata; k++) {           

              inside = 0;
              for (t = 0; t < min_data; t++) 
                   if (k == d_inliers[t])
                      inside = 1;

              if (!inside) {

                  /*ESTA ERRADO O ACESSO AQUI, POIS INLIERS NAO TEM OS PONTOS DE CDAT*/
                  //fprintf(ferased, "thr = %f, point = %d\n", sqrt (tf_norm_calibration_point_error (k, &cdat, inliers_cpar)), k);
                  //if (sqrt (tf_norm_calibration_point_error (k, &inliers_cdat, inliers_cpar)) < threshold) {
                  if (sqrt (tf_norm_calibration_point_error (k, &cdat, inliers_cpar)) < threshold) {
                      consensus_set->p_i[l] = tdat->p_i[k];                          
                      consensus_set->p_w[l] = tdat->p_w[k];                          
                      consensus_set->conf[l] = tdat->conf[k]; 
                      consensus_set->ntargets++;
                      d_consensus[l] = k;
                      l++;
                  }
              }
         } 

         /*This implies that we may have found a good model, now test how good it is*/
         //fprintf(stderr, "inliers = %d,  consensus_set = %d, close_data_values = %d\n", inliers_set->ntargets, consensus_set->ntargets, close_data_values);
         if (consensus_set->ntargets >= close_data_values) { 

              camera_specs_t cspec = tf_get_svga_povray_specs ();
              camera_parameters_t best_cpar = tf_get_mean_camera_parameters (cspec); 

              struct _calibration_data_t best_cdat = (struct _calibration_data_t)
                                                     {
                                                         consensus_set->ntargets,
                                                         consensus_set->p_w,
                                                         consensus_set->p_i,
                                                         consensus_set->conf
                                                     };

              if (! tsaiWeights) {
                tf_calib_set_weights(&best_cdat, consensus_set->conf);
              } else {
                 tf_calib_set_weights_uniform(&best_cdat);
              }
	      tf_calib2_tsai (&best_cdat, cspec, best_cpar, FALSE, TRUE);
//              tf_tsai_calib (&best_cdat, cspec, best_cpar, stderr);  

              double this_mean, this_stddev;
              tf_norm_calibration_error_measures (&best_cdat, best_cpar, &this_mean, &this_stddev);
	      this_error = this_stddev; 

              //fprintf(ferased, "####### this_error = %f, best_error = %f, this_mean %f\n", this_error, best_error, this_mean);

              if (this_error < best_error) { 
	           // fprintf(ferased, "update no melhor conjunto, iter %d\n", iterations);
                   //we have found a model which is better than any of the previous ones
                   //keep it until a better one is found
                   tf_targets_data_copy (consensus_set, best_set); 
                   best_error = this_error; 
                   
                   int t;
                   for (t = 0; t < best_set->ntargets; t++) {
                      d_best[t] = d_consensus[t];
                     //fprintf(ferased, "%d, ", d_best[t]);
		   } 
		   //fprintf(ferased, "\n");
		   found_best = 1;
              }
	      tf_free_camera_parameters_structure (best_cpar);
         } 


         /*Clean-up the choosed points*/
	 //fprintf(ferased, "Inliers: ");

         qsort (d_inliers, min_data, sizeof(int), compare);
	 for (k = 0; k < min_data; k++) {
	     //fprintf(ferased, "%d, ", d_inliers[k]);
	     d_inliers[k] = 0; 
	 }   
         for (k = 0; k < ndata; k++) {
             d_consensus[k] = 0;
         }  
	 //fprintf(ferased, "\n");
         //inliers_set->ntargets = 0;
	 tf_targets_data_clear (consensus_set);
	 tf_targets_data_clear (inliers_set);
         consensus_set->ntargets = 0;

         tf_free_camera_parameters_structure (inliers_cpar);

         iterations++; 
    }

    int k, l;
    for (k = 0; k < ndata; k++) {
        int erase = 1;
        for (l = 0; l < ndata; l++) {
           if (k == d_best[l]) {
               erase = 0;
               fprintf(ferased, "dentro: %d\n", k);
           }
        } 
        if (erase) {
	    if (found_best) {
               tdat->conf[k] = 0.0; 
               fprintf(ferased, "eliminado: %d\n", k);
	    }   
            else
	       tdat->conf[k] = 1.0;
	}
    } 
    fprintf(ferased, "\n");

    //tf_targets_data_free (inliers_set);
    //tf_targets_data_free (consensus_set);
    #undef HORIZONTAL 
    #undef VERTICAL   
}


int compare (const void * a, const void * b)
{
  return ( *(int*)a - *(int*)b );
}

int is_inside (int selected, int d_size, int d_inliers[])
{
    int i;
    for (i = 0; i < d_size; i++) {
        //fprintf(stderr, "%d %d\n", selected, d_size);
        if (selected == d_inliers[i])
            return 1;   
    }
    return 0;
}


