#define _GNU_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <math.h>
#include <image.h>
#include <tf_calib.h>
#include <r2x2.h>
#include <r2.h>
#include "lk.h" 
#include "lk_error_M.h" 
#include "lk_error_STB.h" 
#include "track_parameters.h" 

int main (int argc, char *argv[])
{
  /*Verify memory*/
  void *trash = malloc(1);
  struct mallinfo info;
  int MemDinInicial, MemDinFinal;
  /*End*/

  free(trash);
  info = mallinfo();
  MemDinInicial = info.uordblks;

  image_t I = image_read_pgm (argv[1]);
  image_t M = image_read_pgm (argv[2]);
  image_t W = image_read_pgm (argv[3]);
  
  /* Choose the error function: */
  char *efunc_name = argv[4];
  lk_error_function_t *efunc = parse_error_function_arg(efunc_name);
  printf("error function = %s\n", efunc_name);

  int nz = 2;
  int N = 30;
  double z[nz];

  r2_t p_i;
  p_i.c[0] = M->ny/2.0;
  p_i.c[1] = M->nx/2.0;
  double alpha = 1.0;
  double beta = 0.0;

  auto void goal_func(int nz, double z[], double *Q, double *J, double *H);

  void goal_func(int nz, double z[], double *Q, double *J, double *H)
     { assert(nz == 2);
       r2_t q_i;
       lk_scatter_params (nz, z, &q_i);
       (*efunc)(I, M, W, MARK_SUBSAMPLING, alpha, beta, q_i, Q, J, H);
     }

  track_parameters_print(stderr);

  lk_gather_params (nz, z, p_i);
  double plot_step[nz];
  lk_choose_plot_steps(nz, plot_step);
  double diff_step[nz];
  lk_choose_num_diff_steps(nz, diff_step);
  numeric_plot_all_goal_funcs_and_derivatives ("error", goal_func, nz, z, diff_step, plot_step, N);

  image_free (I);
  image_free (M);
  image_free (W);

  info = mallinfo();
  MemDinFinal = info.uordblks;
  if (MemDinInicial!=MemDinFinal)
    fprintf(stderr, "\n\nMain loop: Dinamic memory was not completely deallocated (%d, %d)\n", MemDinInicial,MemDinFinal);
  return 0;
} 
