#define _GNU_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <math.h>
#include <image.h>
#include <tf_calib.h>
#include <r2x2.h>
#include <r2.h>
#include "lk.h" 
#include "lk_error_M.h" 
#include "track_parameters.h"
#include "lk_error_STB.h" 
#include "numeric.h" 

int main (int argc, char *argv[])
{
  /*Verify memory*/
  void *trash = malloc(1);
  struct mallinfo info;
  int MemDinInicial, MemDinFinal;
  /*End*/

  free(trash);
  info = mallinfo();
  MemDinInicial = info.uordblks;

  image_t I = image_read_pgm (argv[1]);
  image_t M = image_read_pgm (argv[2]);
  image_t W = image_read_pgm (argv[3]);
  
  /* Choose the error function: */
  char *efunc_name = argv[4];
  lk_error_function_t *efunc = parse_error_function_arg(efunc_name);
  printf("error function = %s\n", efunc_name);

  /* Choose the minimizer: */
  char *minize_name = argv[5];
  printf("minimizer = %s\n", minize_name);
  numeric_minimizer_t *minize = parse_minimizer_arg(minize_name);

  /* Choose the minimization step factor: */
  double step_factor = atof(argv[6]);
  printf("step_factor = %8.6f\n", step_factor);

  /* Output filename prefix: */
  char *outname = argv[7];

  /* Choose the central parameters: */
  int nz = 2;
  double z[nz];
  r2_t p_i;
  p_i.c[0] = I->nx/2.0;
  p_i.c[1] = I->ny/2.0;
  lk_gather_params (nz, z, p_i);

  /* Choose the number of plotting steps: */ 
  double alpha = 1.0;
  double beta = 0.0; 
  int niter = 30;
  
  auto void goal_func(int nz, double z[], double *Q, double *J, double *H);

  void goal_func(int nz, double z[], double *Q, double *J, double *H)
     { assert(nz == 2);
       r2_t q_i;
       lk_scatter_params (nz, z, &q_i);
       (*efunc)(I, M, W, MARK_SUBSAMPLING, alpha, beta, q_i, Q, J, H);
     }

  track_parameters_print(stderr);

  lk_gather_params (nz, z, p_i);
  double plot_step[nz];
  lk_choose_plot_steps(nz, plot_step);
  double rplot = 8.0; /* Number of pixels on each side of the center in plot. */ 
  int N = (int)(rplot / fmin(plot_step[0], plot_step[1]));

   /* Max displacement allowed from initial guess [pixels]: */
   /* !!! SHOULD COME FROM A COMMON INTERFACE */
   double max_d = 2.0;
   double max_disp[nz];
   int k;
   for (k = 0; k < nz; k++) { max_disp[k] = max_d; } 
   double min_step = 0.001/max_d; /* Convergence step length (rel to {max_disp}): */

   numeric_plot_2D_goal_func_and_derivatives
    ( outname, 
      nz, z,
      plot_step,  
      goal_func, 
      *minize,
      step_factor,
      max_disp,
      niter,
      min_step,
      N
    );

  image_free (I);
  image_free (M);
  image_free (W);

  info = mallinfo();
  MemDinFinal = info.uordblks;
  if (MemDinInicial!=MemDinFinal)
    fprintf(stderr, "\n\nMain loop: Dinamic memory was not completely deallocated (%d, %d)\n", MemDinInicial,MemDinFinal);
  return 0;
} 
