#define _GNU_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <math.h>
#include <image.h>
#include <tf_calib.h>
#include <r2x2.h>
#include <r2.h>
#include "lk.h" 
#include "lk_error_M.h" 
#include "track_parameters.h"
#include "lk_error_STB.h" 
#include "numeric.h" 

#define NUMBER_OF_BINS 100 

void at_estimate_alpha_beta2
  ( image_t I,
    image_t M, 
    image_t W,
    int subsampling,
    r2_t p_i,
    double *alpha,
    double *beta
  );


int main (int argc, char *argv[])
{
  /*Verify memory*/
  void *trash = malloc(1);
  struct mallinfo info;
  int MemDinInicial, MemDinFinal;
  /*End*/

  free(trash);
  info = mallinfo();
  MemDinInicial = info.uordblks;

  FILE *marks_file = fopen (argv[1], "r");

  /*Initializing the bin vectors*/
  int i; 
  int bins[NUMBER_OF_BINS];
  for (i = 0; i < NUMBER_OF_BINS; i++)
      bins[i] = 0;

  while (1)
  {
      char *N_name, *M_name, *W_name;

      N_name = (char *)malloc(256 * sizeof(char)); 
      M_name = (char *)malloc(256 * sizeof(char)); 
      W_name = (char *)malloc(256 * sizeof(char)); 

      if (fscanf(marks_file, "%s", N_name) != 1) { break; }
      if (fscanf(marks_file, "%s", M_name) != 1) { assert(0);}
      if (fscanf(marks_file, "%s", W_name) != 1) { assert(0); }

      image_t N = image_read_pgm (N_name);
      image_t M = image_read_pgm (M_name);
      image_t W = image_read_pgm (W_name);
 
      printf("reading image %s\n", N_name);
      printf("reading image %s\n", M_name);
      printf("reading image %s\n", W_name);
 
      /* Choose the error function: */
      char *efunc_name = argv[2];
      lk_error_function_t *efunc = parse_error_function_arg(efunc_name);
        printf("error function = %s\n", efunc_name);

      /* Choose the central parameters: */
      int nz = 2;
      double z[nz];
      r2_t p_i;
      p_i.c[0] = N->nx/2.0;
      p_i.c[1] = N->ny/2.0;
      lk_gather_params (nz, z, p_i);

      /* Choose the number of plotting steps: */ 
      //double alpha = 1.0;
      //double beta = 0.0;


      double alpha, beta; 
      at_estimate_alpha_beta2 (N, M, W, MARK_SUBSAMPLING, p_i, &alpha, &beta);
 
      double Q, J[nz], H[nz];

      r2_t q_i;
      lk_scatter_params (nz, z, &q_i);
      (*efunc)(N, M, W, MARK_SUBSAMPLING, alpha, beta, q_i, &Q, J, H);

      printf("cheguei ate aqui\n");

      /*Stores the computed Q in the appropriated bin to compose an histogram*/
      double bin_interval = 1.0/NUMBER_OF_BINS;

      fprintf(stderr, "o bininterval vale %f\n", bin_interval);
      fprintf(stderr, "Q vale %f\n", Q);

      int pos_bin = (int)(sqrt(Q)/bin_interval);

      printf("ate aqui 2\n");
      bins[pos_bin]++;

      printf("ate aqui 3\n");
      image_free (N);
      image_free (M);
      image_free (W);
   }

  char *outPrefix = argv[3];
  
  char *outName = NULL;
  asprintf(&outName, "%s.txt", outPrefix);
  FILE *output = fopen(outName, "w");

   /*Writes the histogram of the bins to a set of images*/
   /*The output are two integer numbers, the bin and the ocurrence of Q that are in that bin*/ 
   double bin_interval = 1.0/NUMBER_OF_BINS;
   for (i = 0; i < NUMBER_OF_BINS; i++)
       fprintf(output, "%f %f %d\n", bin_interval*i, bin_interval*(i+1), bins[i]);   
 
   fclose(output);
   free(outName);

   info = mallinfo();
   MemDinFinal = info.uordblks;
   if (MemDinInicial!=MemDinFinal)
     fprintf(stderr, "\n\nMain loop: Dinamic memory was not completely deallocated (%d, %d)\n", MemDinInicial,MemDinFinal);
   return 0;
}

void at_estimate_alpha_beta2
  ( image_t I,
    image_t M, 
    image_t W,
    int subsampling,
    r2_t p_i,
    double *alpha,
    double *beta
  ) 
{   
  int M_nx = M->nx;
  int M_ny = M->ny;

  /* Compute mean and average of mark in the reference image {M}: */
  r2_t trivial_p_i = (r2_t){{ M_nx/2.0, M_ny/2.0 }};

  double Mavg, Mvar;
  image_weighted_avg_var (M, W, 1, trivial_p_i, &Mavg, &Mvar);
  fprintf(stderr, "Mavg %f Mvar %f\n", Mavg, Mvar);

  /* Compute mean and average of mark in the frame image {I}: */

  double Iavg, Ivar;
  image_weighted_avg_var (I, W, subsampling, p_i, &Iavg, &Ivar);
  fprintf(stderr, "Iavg %f Ivar %f\n", Iavg, Ivar);
    
  *alpha = sqrt (Ivar/Mvar);
  if (*alpha < 0.1) {*alpha = 0.1;}
  if (*alpha > 1.0) {*alpha = 1.0;}
    
  *beta = Iavg - (*alpha)*Mavg;
    
  if (*beta < 0.0) {*beta = 0.0;}
 
  fprintf(stderr, "Estimated alpha = %f beta = %f\n", *alpha, *beta);
}  
