#define _GNU_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <math.h>
#include <image.h>
#include <tf_calib.h>
#include <r2x2.h>
#include <r2.h>
#include "lk.h" 
#include "lk_error_M.h" 
#include "track_parameters.h"
#include "lk_error_STB.h" 
#include "numeric.h" 

int main (int argc, char *argv[])
{
  /*Verify memory*/
  void *trash = malloc(1);
  struct mallinfo info;
  int MemDinInicial, MemDinFinal;
  /*End*/

  free(trash);
  info = mallinfo();
  MemDinInicial = info.uordblks;
 
  /*Assume that the a priori uncertainty is 10 pixels:*/
  r2_t p_i_dev = (r2_t){{ 10.0, 10.0 }};

  image_t I = image_read_pgm (argv[1]);
  image_t M = image_read_pgm (argv[2]);
  image_t W = image_read_pgm (argv[3]);
  
  /* Choose the error function: */
  char *efunc_name = argv[4];
  lk_error_function_t *efunc = parse_error_function_arg(efunc_name);
  printf("error function = %s\n", efunc_name);

  /* Choose the minimizer: */
  char *minize_name = argv[5];
  printf("minimizer = %s\n", minize_name);
  numeric_minimizer_t *minize = parse_minimizer_arg(minize_name);

  /* Choose the minimization step factor: */
  double step_factor = atof(argv[6]);
  printf("step_factor = %8.6f\n", step_factor);

  

  double alpha = atof(argv[7]);
  double beta  = atof(argv[8]);
  r2_t p_i;
  p_i.c[0] = atof(argv[9]);
  p_i.c[1] = atof(argv[10]);
  int niter = atoi(argv[11]);

  int subsampling = 1;

  /* Output filename prefix: */
  char *outname = argv[12];

     auto void goal_func(int nz, double z[], double *Q, double *J, double *H);

     void goal_func(int nz, double z[], double *Q, double *J, double *H)
     { assert(nz == 2);
       r2_t q_i;
       lk_scatter_params (nz, z, &q_i);
       (*efunc)(I, M, W, MARK_SUBSAMPLING, alpha, beta, q_i, Q, J, H);
     }

    track_parameters_print(stderr);

    /*Ploting*/
    
    int nz = 2;
    double z[nz];
    lk_gather_params (nz, z, p_i);
    double plot_step[nz];
    lk_choose_plot_steps(nz, plot_step);
    double rplot = 5.0; /* Number of pixels on each side of the center in plot. */
    int N = (int)(rplot / fmin(plot_step[0], plot_step[1]));

    double max_d = 2.0;
    double max_disp[nz];
    int k;
    for (k = 0; k < nz; k++) { max_disp[k] = max_d; }
    double min_step = 0.001/max_d; /* Convergence step length (rel to {max_disp}): */

    numeric_plot_2D_goal_func_and_derivatives
    ( outname,
      nz, z,
      plot_step,
      goal_func,
      *minize,
      step_factor,
      max_disp,
      niter,
      min_step,
      N
    ); 

    lk_write_mark_extracted_from_image (I, M->nx, M->ny, subsampling, p_i, alpha, beta, "LK", "TESTAO.pgm");


    /*End ploting*/

    double errorSqr;
    char *fname = "out_name";

    lk_mark_position_adjust
    (   fname,
        I,
        M,
        W,
        subsampling,
        alpha,
        beta,
        &p_i,
        &p_i_dev,
        efunc,
        minize,
        step_factor,
        niter,
        &errorSqr
    ); 

    fprintf(stderr, "Sqr error: %f, x = %f, y = %f\n", errorSqr, p_i.c[0], p_i.c[1]);

    image_free (I);
    image_free (M);
    image_free (W);

    info = mallinfo();
    MemDinFinal = info.uordblks;
    if (MemDinInicial!=MemDinFinal)
       fprintf(stderr, "\n\nMain loop: Dinamic memory was not completely deallocated (%d, %d)\n", MemDinInicial,MemDinFinal);
    return 0;
} 
