#include "cimage.h"

cimage_t cimage_alloc (int nx, int ny, int ch) 
{
    int i;
    /* Allocates image struct */
    cimage_t cimg = (cimage_t)malloc(sizeof(struct _cimage_t) );
    
    cimg->nx = nx;
    cimg->ny = ny;
    cimg->ch = ch;

    cimg->C = (image_t *)malloc(ch * sizeof(image_t));

    /* Allocates bidimensional image arrays for all the image channels */
    for (i = 0; i < ch; i++) {
       cimg->C[i] = image_alloc (nx, ny);
       if (cimg->C[i] == NULL) {
           fprintf( stderr, "error: allocating cimage array" );
           return NULL;
       }
    }   
    return cimg;      /* Returns image struct */
}

void cimage_free (cimage_t cimg)
{
    int i;
    for (i = 0; i < cimg->ch; i++) {
        image_free (cimg->C[i]);
    }
    free(cimg->C);
    free(cimg);
}


void cimage_write_to_ppm (cimage_t cimg, int magic, int maxval, char *fname)
{
    int row, col, ch;
    FILE *fp;
    int bytes;
    unsigned char temp[2];
    
    if ( (fp = fopen(fname, "w")) == NULL)
       image_message_error ("(%s) Can't open file named '%s' for reading\n", __FUNCTION__, fname);

    fprintf(fp, "P%d\n",   magic);
    fprintf(fp, "%d %d\n", cimg->nx, cimg->ny);
    fprintf(fp, "%d\n",    maxval);

    if (maxval <= 255)
         bytes = 1;
    else
         bytes = 2;


    for ( row = 0; row < cimg->ny; row++ ) {
        for ( col = 0; col < cimg->nx; col++ ) {
           double fval, sval;
           for (ch = 0; ch < cimg->ch; ch++) {
                fval = cimg->C[ch]->val[row][col];
                fval = pow(fval, 1.0/GAMMA);
                sval = fval * maxval;
                if (sval > maxval) sval = maxval;
                if (sval < 0) sval = 0;
                if (bytes == 1)
                   temp[0] = (unsigned char)sval;
                else {
                   unsigned short v = (unsigned short)sval;
                   temp[0] = v >> 8;
                   temp[1] = v & 255;
                }
                fwrite( &temp, bytes, 1, fp );
	    }   
        }
    }
    fclose(fp);
}

 

