#include "convert.h"

void convert_RGB_to_YCbCr (cimage_t cimg)
{
   assert(cimg->ch == 3);

   int iy, ix;
   double max = -HUGE, min = HUGE; 
   for (iy = 0; iy < cimg->ny; iy++) {
      for (ix = 0; ix < cimg->nx; ix++) {
         double R = cimg->C[0]->val[iy][ix];
         double G = cimg->C[1]->val[iy][ix];
         double B = cimg->C[2]->val[iy][ix];

         double Y  = +0.298911*R +0.586611*G +0.114478*B;
         double Cb = -0.168777*R -0.331223*G +0.500000*B;
         double Cr = +0.500000*R -0.418357*G -0.081643*B;
 
         /*Normalizing the RGB values to 0-1*/
         double fval = (Y + 0.5)/MAXVAL;
         cimg->C[0]->val[iy][ix] = pow(fval, GAMMA);
         cimg->C[1]->val[iy][ix] = Cb/MAXVAL;
         cimg->C[2]->val[iy][ix] = Cr/MAXVAL;

         if (Y < min)
	    min = Y;
	 if (Y > max)
	    max = Y;
      }
   }
   fprintf(stderr, "Menor Y = %f, maior = %f\n", min, max);
}


image_t convert_Y_cimg_to_img (cimage_t cimg)
{
   image_t out = image_alloc (cimg->nx, cimg->ny); 
  
   int iy, ix;
   for (iy = 0; iy < cimg->ny; iy++) {
      for (ix = 0; ix < cimg->nx; ix++) {
         /*Be carefull - only to YCbCr images*/
         out->val[iy][ix] = cimg->C[0]->val[iy][ix];
      }
   }
   return out; 
}

unsigned char * convert_cimg_to_unsigned_char (cimage_t cimg)
{
   unsigned char *uimg, *ptr;

   uimg = (unsigned char *)malloc(cimg->nx*cimg->ny*sizeof(char));

   ptr = uimg;

   int iy, ix;
   
   for (iy = 0; iy < cimg->ny; iy++) {
      for (ix = 0; ix < cimg->nx; ix++) {
          *ptr = cimg->C[0]->val[iy][ix];
          ptr++;
      }
   }
   return uimg;
}

unsigned char * convert_img_to_unsigned_char (image_t img)
{
   unsigned char *uimg, *ptr;

   uimg = (unsigned char *)malloc(img->nx*img->ny*sizeof(unsigned char));

   ptr = uimg;

   int iy, ix;
   
   for (iy = 0; iy < img->ny; iy++) {
      for (ix = 0; ix < img->nx; ix++) {
          *ptr = img->val[iy][ix];
          ptr++;
      }
   }
   return uimg;
}

image_t convert_unsigned_char_to_img (unsigned char *uimg, int nx, int ny)
{
   unsigned char *ptr;

   ptr = uimg;

   image_t img = image_alloc (nx, ny);

   int iy, ix;
   
   for (iy = 0; iy < ny; iy++) {
      for (ix = 0; ix < nx; ix++) {
          img->val[iy][ix] = *ptr;
          ptr++;
      }
   }
   return img;
}

image_t convert_unsigned_char_to_cimg (unsigned char *uimg, int nx, int ny)
{
   unsigned char *ptr;

   ptr = uimg;

   image_t img = image_alloc (nx, ny);

   int iy, ix;
   
   for (iy = 0; iy < ny; iy++) {
      for (ix = 0; ix < nx; ix++) {
          img->val[iy][ix] = *ptr;
          ptr++;
      }
   }
   return img;
}

void convert_copy_img_to_unsigned_char (image_t img, unsigned char *uimg)
{
   unsigned char *ptr;

   ptr = uimg;

   int iy, ix;
   
   for (iy = 0; iy < img->ny; iy++) {
      for (ix = 0; ix < img->nx; ix++) {
          *ptr = img->val[iy][ix];
          ptr++;
      }
   }
}


void convert_copy_unsigned_char_to_img (unsigned char *uimg, image_t img)
{
   unsigned char *ptr;

   ptr = uimg;

   int iy, ix;
   
   for (iy = 0; iy < img->ny; iy++) {
      for (ix = 0; ix < img->nx; ix++) {
          img->val[iy][ix] = *ptr;
          ptr++;
      }
   }
}

void convert_copy_unsigned_char_to_cimg (unsigned char *uimg, cimage_t cimg)
{
   unsigned char *ptr;

   ptr = uimg;

   int iy, ix;
   
   for (iy = 0; iy < cimg->ny; iy++) {
      for (ix = 0; ix < cimg->nx; ix++) {
          cimg->C[0]->val[iy][ix] = *ptr;
          ptr++;
      }
   }
}











