#include "ijpeg.h"

cimage_t ijpeg_read_jpeg (char *filename)
{
    cimage_t cimg;
    FILE *file;

    /* The "b" option is for binary input, which is needed if this is
       compiled under Windows.  It has no effect in Linux.
    */
    file = fopen (filename, "rb");
    if (!file) 
        ijpeg_fatal_error ("Could not open file: %s", filename);

    cimg = read_jpeg (file);
    fclose(file);
    
    return cimg;
}

cimage_t read_jpeg (FILE *fp)
{
  cimage_t cimg;
  struct jpeg_decompress_struct cinfo;
  struct jpeg_error_mgr jerr;
  JSAMPARRAY buffer;
  int row_stride;
  int r, c;

  cinfo.err = jpeg_std_error(&jerr);
  jpeg_create_decompress(&cinfo);

  jpeg_stdio_src(&cinfo, fp);

  (void) jpeg_read_header(&cinfo, True);
  (void) jpeg_start_decompress(&cinfo);

  row_stride = cinfo.output_width * cinfo.output_components;
  buffer = (*cinfo.mem->alloc_sarray)
                ((j_common_ptr) &cinfo, JPOOL_IMAGE, row_stride, 1);

  int channels = 3;
  cimg = cimage_alloc (cinfo.output_width, cinfo.output_height, channels);
  r = 0;
  while (cinfo.output_scanline < cinfo.output_height) {
    (void) jpeg_read_scanlines(&cinfo, buffer, 1);
    for (c = 0; c < cinfo.output_width; c++) {
      cimg->C[0]->val[r][c] = (int) buffer[0][c*cinfo.output_components+0];
      cimg->C[1]->val[r][c] = (int) buffer[0][c*cinfo.output_components+1];
      cimg->C[2]->val[r][c] = (int) buffer[0][c*cinfo.output_components+2];
    }
    r++;
  }
  (void) jpeg_finish_decompress(&cinfo);
  jpeg_destroy_decompress(&cinfo);

  return cimg;
}


void ijpeg_write_jpeg (char *filename, cimage_t cimg)
{
    FILE *file;

    /* The "b" option is for binary input, which is needed if this is
       compiled under Windows.  It has no effect in Linux.
    */
    file = fopen (filename, "wb");
    if (! file)
        ijpeg_fatal_error ("Could not open file: %s", filename);

    write_jpeg (file, cimg);
    fclose(file);
}

void write_jpeg (FILE *fp, cimage_t cimg)
{
  struct jpeg_compress_struct cinfo;
  struct jpeg_error_mgr jerr;
  JSAMPARRAY buffer;
  int row_stride;
  int r, c;

  assert(cimg->ch == 3);

  cinfo.err = jpeg_std_error(&jerr);
  jpeg_create_compress(&cinfo);

  jpeg_stdio_dest(&cinfo, fp);

  cinfo.image_width = cimg->nx;
  cinfo.image_height = cimg->ny;
  cinfo.input_components = 3;
  cinfo.in_color_space = JCS_RGB;

  jpeg_set_defaults(&cinfo);

  jpeg_start_compress(&cinfo, True);

  row_stride = cinfo.image_width * cinfo.input_components;
  buffer = (*cinfo.mem->alloc_sarray)
                ((j_common_ptr) &cinfo, JPOOL_IMAGE, row_stride, 1);

  r = 0;
  while (cinfo.next_scanline < cinfo.image_height) {
    for (c = 0; c < cinfo.image_width; c++) {
      buffer[0][c*cinfo.input_components+0] = (int) (cimg->C[0]->val[r][c]);
      buffer[0][c*cinfo.input_components+1] = (int) (cimg->C[1]->val[r][c]);
      buffer[0][c*cinfo.input_components+2] = (int) (cimg->C[2]->val[r][c]);
    }
    r++;
    (void) jpeg_write_scanlines(&cinfo, buffer, 1);
  }
  jpeg_finish_compress(&cinfo);
  jpeg_destroy_compress(&cinfo);
}

void ijpeg_fatal_error (char *fmt, ...)
{
    va_list args;

    va_start(args, fmt);
    fprintf(stderr, "Error: ");
    vfprintf(stderr, fmt, args);
    fprintf(stderr,"\n");
    va_end(args);
    exit(1);
}
