#include "image.h"
#include <jsfile.h>
#include <affirm.h>

#define DEBUG_ALL 0

image_t image_copy (image_t img)
{
    int iy, ix;

    /*new image*/
    image_t nimg;

    nimg = image_alloc (img->nx, img->ny);

    for (iy = 0; iy < img->ny; iy++) {
        for (ix = 0; ix < img->nx; ix++) {
            nimg->val[iy][ix] = img->val[iy][ix]; 
        }
    }
    return nimg;
}

image_t image_alloc (int nx, int ny) 
{
  
  /* Allocates image struct */
    image_t img = (image_t)malloc(sizeof(struct _image_t) );
    
    img->nx = nx;
    img->ny = ny;

    /* Allocates bidimensional array for image pixels. Note reverse order of args! */
    img->val = image_mdouble_alloc (ny, nx);
    return img;      /* Returns image struct */
}

void image_free (image_t img)
{
    image_mdouble_free (img->val, img->ny);
    free(img);
}

image_t image_read_pgm (char *fname)
{
    int row, col, ny, nx;
    int magic, maxval;
    int bytes;
    image_t img;
    unsigned char temp[2];
    FILE *fp;
    
    /* Allocates image struct */
    img = (image_t)malloc(sizeof(struct _image_t));
    
    /* Open file */
    if ( (fp = fopen(fname, "rb")) == NULL)
       image_message_error ("(read_pgm_image) can't open file named '%s' for reading\n", fname);

    /* Read header */
    image_read_pgm_header (fp, &magic, &nx, &ny, &maxval);
    
    img->ny = ny;
    img->nx = nx;

    if (maxval <= 255)
         bytes = 1;
    else
         bytes = 2;        

    /* Allocates bidimensional array for image pixels */
    img->val = image_mdouble_alloc (ny, nx);

    for ( row = 0; row < img->ny; row++ ) {
        for ( col = 0; col < img->nx; col++ ) {

             fread( &temp, bytes, 1, fp );
             int ival;
	     if (bytes == 1)
                 ival = temp[0];
             else
                 ival = ((int)(temp[0]))*256 + (int)(temp[1]);
             double fval = ((double)ival + 0.5)/maxval;
             fval = pow(fval, GAMMA);
             img->val[row][col] = fval;
        }
    }
    fclose(fp);
    return img;      /* Returns image struct */
}

void image_write_pgm (image_t img, int magic, int maxval, char *fname)
{
    int row, col;
    FILE *fp = open_write(fname, TRUE);
    int bytes;
    unsigned char temp[2];
  
    fprintf(fp, "P%d\n",   magic);
    fprintf(fp, "%d %d\n", img->nx, img->ny);
    fprintf(fp, "%d\n",    maxval);

    if (maxval <= 255)
         bytes = 1;
    else
         bytes = 2;

    for ( row = 0; row < img->ny; row++ ) {
        for ( col = 0; col < img->nx; col++ ) {
	  double fval = img->val[row][col];
          fval = pow(fval, 1.0/GAMMA);
          double sval = fval * maxval;
          if (sval > maxval) sval = maxval;
          if (sval < 0) sval = 0;
          if (bytes == 1)
              temp[0] = (unsigned char)sval;
          else {
             unsigned short v = (unsigned short)sval;
             temp[0] = v >> 8;
             temp[1] = v & 255;
          }
          fwrite( &temp, bytes, 1, fp );
        }
    }
    fclose(fp);
}

void image_print_cross (image_t img, int size, r2_t pi)
{
   int col = (int)(pi.c[0]);
   int row = (int)(pi.c[1]);

   int blacksize = (size < 3 ? 1 : size-2);
   int whitesize = size;
  
   int dx, dy;
   /* Paint the black part: */
   for (dy = - blacksize ; dy <= + blacksize ; dy++) {
      for (dx = - blacksize ; dx <= + blacksize ; dx++) {
        bool_t h_cross_frame = (abs(dy) == 1) && (abs(dx) <= blacksize-1);
	bool_t v_cross_frame = (abs(dx) == 1) && (abs(dy) <= blacksize-1);
	if (h_cross_frame || v_cross_frame) {
	  int xx = col + dx;
	  int yy = row + dy;
	  img->val[yy][xx] = 0;
         }
      }
   }
   /* Paint the white part: */
   for (dy = - whitesize ; dy <= + whitesize ; dy++) {
      for (dx = - whitesize ; dx <= + whitesize ; dx++) {
	bool_t h_cross = dy == 0;
	bool_t v_cross = dx == 0;
	if (h_cross || v_cross) {
	  int xx = col + dx;
	  int yy = row + dy;
	  img->val[yy][xx] = 1;
         }
      }
   }
}

void image_print_dot (image_t img, int size, r2_t pi)
{
   int col = (int)(pi.c[0]);
   int row = (int)(pi.c[1]);
  
   int dx, dy;
   /* Paint the black part: */
   for (dy = - size ; dy <= + size ; dy++) {
      for (dx = - size ; dx <= + size ; dx++) {
        bool_t outer_frame = (abs(dx) == size-1) || (abs(dy) == size-1);
	bool_t inner_frame = (abs(dx) <= 1) && (abs(dy) <= 1);
	if (outer_frame || inner_frame) {
	  int xx = col + dx;
	  int yy = row + dy;
	  img->val[yy][xx] = 0;
         }
      }
   }
   /* Paint the white part: */
   for (dy = - size ; dy <= + size ; dy++) {
      for (dx = - size ; dx <= + size ; dx++) {
	bool_t outer_frame = (abs(dx) == size) || (abs(dy) == size);
	bool_t inner_frame = (abs(dx) == 0) && (abs(dy) == 0);
	if (outer_frame || inner_frame) {
	  int xx = col + dx;
	  int yy = row + dy;
	  img->val[yy][xx] = 1;
         }
      }
   }
}

void image_ppm_write_targets_with_crosses 
  ( image_t img, r2_t *p_i, int ntargets, int magic, int maxval, int magnify, char *fname)
{
    int row, col;
    FILE *fp = open_write(fname, TRUE);
    int bytes;
    unsigned char temp[2];
    image_t redimg;
    image_t blueimg;
    image_t greenimg;

    int bigrows = img->ny*magnify;
    int bigcols = img->nx*magnify; 
  
    fprintf(fp, "P%d\n",   magic);
    fprintf(fp, "%d %d\n", bigcols, bigrows);
    fprintf(fp, "%d\n",    maxval);

    if (maxval <= 255)
         bytes = 1;
    else
         bytes = 2;

    redimg = image_alloc(bigcols, bigrows);
    blueimg = image_alloc(bigcols, bigrows);
    greenimg = image_alloc(bigcols, bigrows);
    int x, y;
    double oldx, oldy;
    for (y = 0; y < bigrows; y++)
    for (x = 0; x < bigcols; x++) {
      double v;
      oldx = ((double)x + 0.5)/((double)magnify);
      oldy = ((double)y + 0.5)/((double)magnify);
      image_eval_bilinear (oldx, oldy, img, &v); 
      redimg->val[y][x] = v;
      blueimg->val[y][x] = v;
      greenimg->val[y][x] = v;
    }

    int i; 
    for (i = 0; i < ntargets; i++) {
        int mark_radius = 4;
        int hextrows = (int)(p_i[i].c[1]*magnify);
        int hextcols = (int)(p_i[i].c[0]*magnify);

        int xx, yy;
        for (yy = hextrows - mark_radius ; yy <= hextrows + mark_radius ; yy++) {
           for (xx = hextcols - mark_radius ; xx <= hextcols + mark_radius ; xx++) {
              if ((xx == hextcols) || (yy == hextrows)) { 
	         if ((xx >= 0) && (xx < bigcols-1) && (yy >= 0) && (yy < bigrows-1)) {
                    redimg->val[yy][xx] = 1;
                    blueimg->val[yy][xx] = 1;
                    greenimg->val[yy][xx] = 1;
                    redimg->val[yy+1][xx+1] = 0;
                    blueimg->val[yy+1][xx+1] = 0;
                    greenimg->val[yy+1][xx+1] = 0;
		  }
              }
	   }
	}   
    }

    for ( row = 0; row < bigrows; row++ ) {
        for ( col = 0; col < bigcols; col++ ) {
          double fval, sval;
     
	  fval = redimg->val[row][col];
          fval = pow(fval, 1.0/GAMMA);
          sval = fval * maxval;
          if (sval > maxval) sval = maxval;
          if (sval < 0) sval = 0;
          if (bytes == 1)
              temp[0] = (unsigned char)sval;
          else {
             unsigned short v = (unsigned short)sval;
             temp[0] = v >> 8;
             temp[1] = v & 255;
          }
          fwrite( &temp, bytes, 1, fp );

          fval = blueimg->val[row][col];
          fval = pow(fval, 1.0/GAMMA);
          sval = fval * maxval;
          if (sval > maxval) sval = maxval;
          if (sval < 0) sval = 0;
          if (bytes == 1)
              temp[0] = (unsigned char)sval;
          else {
             unsigned short v = (unsigned short)sval;
             temp[0] = v >> 8;
             temp[1] = v & 255;
          }
          fwrite( &temp, bytes, 1, fp );

          fval = greenimg->val[row][col];
          fval = pow(fval, 1.0/GAMMA);
          sval = fval * maxval;
          if (sval > maxval) sval = maxval;
          if (sval < 0) sval = 0;
          if (bytes == 1)
              temp[0] = (unsigned char)sval;
          else {
             unsigned short v = (unsigned short)sval;
             temp[0] = v >> 8;
             temp[1] = v & 255;
          }
          fwrite( &temp, bytes, 1, fp );

        }
    }
    image_free(redimg);
    image_free(blueimg);
    image_free(greenimg);
    fclose(fp);
}

void image_read_pgm_header (FILE *fp, int *magic, int *nx, int *ny, int *maxval)
{
   char line[LENGTH];

   /* Read magic number */
   image_get_next_string (fp, line);
   if (line[0] != 'P') {
       image_message_error ("(pgm read header) magic number does not begin with 'P', but with a '%c'", line[0]);
   }
   sscanf(line, "P%d", magic);

   if (*magic != 5)
      image_message_error ("(pgmReadHeader) Magic number is not 'P5', but 'P%d'", *magic);

   /* Read size, skipping comments */
   image_get_next_string (fp, line);
   *nx = atoi(line);
   image_get_next_string (fp, line);
   *ny = atoi(line);
   if (*nx < 0 || *ny < 0 || *nx > 10000 || *ny > 10000)
       image_message_error ("(pnmReadHeader) The dimensions %d x %d are unacceptable", *nx, *ny);  

   /* Read maxval, skipping comments */
   image_get_next_string (fp, line);
   *maxval = atoi(line);
   fread(line, 1, 1, fp); /* Read newline which follows maxval */ 
}

void image_get_next_string (FILE *fp, char *line)
{ 
   int i;

   line[0] = '\0';
  
   while (line[0] == '\0')  {
      fscanf(fp, "%s", line);
      i = -1;
      do  {
        i++;
        if (line[i] == '#')  {
           line[i] = '\0';
           while (fgetc(fp) != '\n') ;
        }
      }  while (line[i] != '\0');
   }
}


void image_change_interval (image_t img, double vmax, double vmin, double min, double max)
{   
    int iy, ix;

    if (vmax == vmin) {
        vmax = -HUGE_VAL; vmin = +HUGE_VAL;

        for (iy = 0; iy < img->ny; iy++) {
            for (ix = 0; ix < img->nx; ix++) {
               vmax = fmax(vmax, img->val[iy][ix]);
               vmin = fmin(vmin, img->val[iy][ix]);
            }
        }
    }
    if (DEBUG_ALL) {
      fprintf(stderr, "mapping [ %24.16e _ %24.16e ] -> [ %24.16e _ %24.16e ]\n", vmin, vmax, min, max);
    }

    for (iy = 0; iy < img->ny; iy++) {
        for (ix = 0; ix < img->nx; ix++) {
           img->val[iy][ix] = ( (max - min)*(img->val[iy][ix] - vmin) )/(vmax - vmin) + min;
        }
    }
}

void image_change_interval_and_write
  ( image_t img, 
    int magic, 
    int maxval, 
    char *iname, 
    char *type, 
    double vmin,
    double vmax, 
    double min, 
    double max
  )
{
  image_t tmp = image_copy (img);
  if (min != max) { image_change_interval (tmp, vmin, vmax, min, max); }
  char *fname = NULL;
  asprintf(&fname,"%s.%s", iname, type);
  image_write_pgm (tmp, magic, maxval, fname);
  free(fname);
  image_free (tmp);
}

void image_eval_bilinear
  ( double x,
    double y,
    image_t img,
    double *v
  )
{
    x -= 0.5;
    y -= 0.5;

    int nx = img->nx;
    int ny = img->ny;

    int ix = floor(x);
    int iy = floor(y);
    double xfrac = x - ix;
    double yfrac = y - iy;
    double p00, p10, p01, p11;
    double w00, w10, w01, w11;

    if (((ix+0) >= 0) && ((ix+0) < nx) && ((iy+0) >= 0) && ((iy+0) < ny)) {
       p00 = img->val[iy+0][ix+0];
       w00 = 1.0;
    }
    else {
      p00 = w00 = 0.0;
    }

    if (((ix+1) >= 0) && ((ix+1) < nx) && ((iy+0) >= 0) && ((iy+0) < ny)) {
       p10 = img->val[iy+0][ix+1];
       w10 = 1.0;
    }
    else {
      p10 = w10 = 0.00;
    }

    if (((ix+0) >= 0) && ((ix+0) < nx) && ((iy+1) >= 0) && ((iy+1) < ny)) {
       p01 = img->val[iy+1][ix+0];
       w01 = 1.0;
    }
    else {
      p01 = w01 = 0.00;
    }
    
    if (((ix+1) >= 0) && ((ix+1) < nx) && ((iy+1) >= 0) && ((iy+1) < ny)) {
       p11 = img->val[iy+1][ix+1];
       w11 = 1.0;
    }
    else {
      p11 = w11 = 0.00;
    }

    interpol_bilinear (xfrac, yfrac, p00, w00, p10, w10, p01, w01, p11, w11, v);

}


void image_eval_3_A
  ( double x,
    double y,
    image_t img,
    double *v
  )
{
    /* Adjust (x,y) for the (0.5,0.5) pixel center shift: */
    x -= 0.5;
    y -= 0.5;

    int nx = img->nx;
    int ny = img->ny;

    /* Bicubic interpolation: */
    int ix = floor(x);
    int iy = floor(y);
    double xfrac = x - ix;
    double yfrac = y - iy;

    int jx;
    double t[4];
    for(jx = ix-1; jx < (ix+3); jx++){
      int ind = jx - ix +1;
      if ((jx >= 0) && (jx < nx)) {
        double p[4];
        p[0] = (((iy-1) >= 0) && ((iy-1) < ny) ? img->val[iy-1][jx]:0);
        p[1] = (((iy+0) >= 0) && ((iy+0) < ny) ? img->val[iy+0][jx]:0);
        p[2] = (((iy+1) >= 0) && ((iy+1) < ny) ? img->val[iy+1][jx]:0);
        p[3] = (((iy+2) >= 0) && ((iy+2) < ny) ? img->val[iy+2][jx]:0);
        interpol_spline_3_A (yfrac, p, &(t[ind]), NULL, NULL);
      } else {
        t[ind] = 0.0;
      }
    }
    interpol_spline_3_A (xfrac, t, v, NULL, NULL);
}

void image_eval_3_B
  ( double x,
    double y,
    image_t img,
    double *v
  )
{
    /* Adjust (x,y) for the (0.5,0.5) pixel center shift: */
    x -= 0.5;
    y -= 0.5;

    int nx = img->nx;
    int ny = img->ny;

    /* Bicubic interpolation: */
    int ix = floor(x);
    int iy = floor(y);
    double xfrac = x - ix;
    double yfrac = y - iy;

    int jx;
    double t[4];
    for(jx = ix-1; jx < (ix+3); jx++){
      int ind = jx - ix +1;
      if ((jx >= 0) && (jx < nx)) {
        double p[4];
        p[0] = (((iy-1) >= 0) && ((iy-1) < ny) ? img->val[iy-1][jx]:0);
        p[1] = (((iy+0) >= 0) && ((iy+0) < ny) ? img->val[iy+0][jx]:0);
        p[2] = (((iy+1) >= 0) && ((iy+1) < ny) ? img->val[iy+1][jx]:0);
        p[3] = (((iy+2) >= 0) && ((iy+2) < ny) ? img->val[iy+2][jx]:0);
        interpol_spline_3_B (yfrac, p, &(t[ind]), NULL, NULL);
      } else {
        t[ind] = 0.0;
      }
    }
    interpol_spline_3_B (xfrac, t, v, NULL, NULL);
}

void image_eval_3_C
  ( double x,
    double y,
    image_t img,
    double *v
  )
{
    /* Adjust (x,y) for the (0.5,0.5) pixel center shift: */
    x -= 0.5;
    y -= 0.5;

    int nx = img->nx;
    int ny = img->ny;

    /* Bicubic interpolation: */
    int ix = floor(x);
    int iy = floor(y);
    double xfrac = x - ix;
    double yfrac = y - iy;

    int jx;
    double t[4];
    for(jx = ix-1; jx < (ix+3); jx++){
      int ind = jx - ix +1;
      if ((jx >= 0) && (jx < nx)) {
        double p[4];
        p[0] = (((iy-1) >= 0) && ((iy-1) < ny) ? img->val[iy-1][jx]:0);
        p[1] = (((iy+0) >= 0) && ((iy+0) < ny) ? img->val[iy+0][jx]:0);
        p[2] = (((iy+1) >= 0) && ((iy+1) < ny) ? img->val[iy+1][jx]:0);
        p[3] = (((iy+2) >= 0) && ((iy+2) < ny) ? img->val[iy+2][jx]:0);
        interpol_spline_3_C (yfrac, p, &(t[ind]), NULL, NULL);
      } else {
        t[ind] = 0.0;
      }
    }
    interpol_spline_3_C (xfrac, t, v, NULL, NULL);
}

void image_message_error (char *fmt, ...)
{
    va_list args;
    va_start(args, fmt);
    fprintf(stderr, "error: ");
    vfprintf(stderr, fmt, args);
    fprintf(stderr, "\n");
    va_end(args);
    exit(1);
}

image_t image_diff_x_direction (image_t img)
{
    int iy, ix;
    image_t der;

    der = image_alloc(img->nx, img->ny);

    for (iy = 1; iy < (img->ny - 1); iy++) {
        for (ix = 1; ix < (img->nx - 1); ix++) {
            der->val[iy][ix] = (img->val[iy][ix+1] - img->val[iy][ix-1])/2.0;
        }
    }
    return der;
}

image_t image_diff_y_direction (image_t img)
{
    int iy, ix;
    image_t der;

    der = image_alloc(img->nx, img->ny);

    for (iy = 1; iy < (img->ny - 1); iy++) {
        for (ix = 1; ix < (img->nx - 1); ix++) {
            der->val[iy][ix] = (img->val[iy+1][ix] - img->val[iy-1][ix])/2.0;
        }
    }
    return der;
}

image_t image_diff_xx_direction (image_t img)
{
    int iy, ix;
    image_t der;
    
    der = image_alloc(img->nx, img->ny);

    for (iy = 1; iy < (img->ny - 1); iy++) {
        for (ix = 1; ix < (img->nx - 1); ix++) {
            der->val[iy][ix] = img->val[iy][ix+1] -2 * img->val[iy][ix] + img->val[iy][ix-1];
        }
    }
    return der;
}

image_t image_diff_xy_direction (image_t img)
{
    int iy, ix;
    image_t der;
    
    der = image_alloc(img->nx, img->ny);

    for (iy = 1; iy < (img->ny - 1); iy++) {
        for (ix = 1; ix < (img->nx - 1); ix++) {
            der->val[iy][ix] = (img->val[iy+1][ix+1] - img->val[iy-1][ix+1] - img->val[iy+1][ix-1] + img->val[iy-1][ix-1])/4.0;
        }
    }
    return der;
}

image_t image_diff_yy_direction (image_t img)
{
    int iy, ix;
    image_t der;
    
    der = image_alloc(img->nx, img->ny);

    for (iy = 1; iy < (img->ny - 1); iy++) {
        for (ix = 1; ix < (img->nx - 1); ix++) {
            der->val[iy][ix] = img->val[iy+1][ix] -2 * img->val[iy][ix] + img->val[iy-1][ix];
        }
    }
    return der;
}

image_t image_gaussian_mask (int nx, int ny, int normalize)
{
    int iy, ix;
    double sum = 0.0;
    double cte = 0.4;
    double x_center = nx/2.0;
    double y_center = ny/2.0;
    double sigma_x = cte * (nx/2.0);
    double sigma_y = cte * (ny/2.0);

    image_t gw = image_alloc (nx, ny);

    for (iy = 0; iy < ny; iy++) {
        for (ix = 0; ix < nx; ix++) {
            double x, y; /*coordinates of pixel center*/

            image_pixel_center (ix, iy, &x, &y); 

            double dr = ( x - x_center ) / sigma_x;
            double dc = ( y - y_center ) / sigma_y;

            gw->val[iy][ix] = exp (-(dr * dr + dc * dc)/2.0);

            sum += gw->val[iy][ix];
            if (DEBUG_ALL) {
               printf("iy %d ix %d dr %f dc %f ny %d nx %d val = %f \n", iy, ix, dr, dc, ny, nx, gw->val[iy][ix]);
	    }
        }
    }
    if (normalize) {
        for (iy = 0; iy < ny; iy++) {
            for (ix = 0; ix < nx; ix++) {
                gw->val[iy][ix] /= sum;
            }
        }
    }
    return gw;
}

void image_pixel_center (int ix, int iy, double *x, double *y)
{
   *x = ix + 0.5; 
   *y = iy + 0.5; 
}

void image_map_coords (double ax, double ay, r3x3_t *map, double *bx, double *by)
{
  if ( (map->c[0][0] != 1) || (map->c[0][1] != 0) || (map->c[0][2] != 0) ) {
    fprintf (stderr,  "map0 = %f, %f, %f\n", map->c[0][0], map->c[0][1], map->c[0][2]);
    assert(FALSE);
  }
  *bx = map->c[1][0] + map->c[1][1]*ax + map->c[1][2]*ay;
  *by = map->c[2][0] + map->c[2][1]*ax + map->c[2][2]*ay;
}

void image_map_mark_coords_to_image_coords (double ax, double ay, int Mnx, int Mny, int subsampling, r2_t p_i, double *bx, double *by)
{
  *bx = (ax - Mnx/2.0)/subsampling + p_i.c[0];
  *by = (ay - Mny/2.0)/subsampling + p_i.c[1];
}

void image_map_image_coords_to_mark_coords (double bx, double by, int Mnx, int Mny, int subsampling, r2_t p_i, double *ax, double *ay)
{
  *ax = (bx - p_i.c[0])*subsampling + Mnx/2.0;
  *ay = (by - p_i.c[1])*subsampling + Mny/2.0;
}

void image_map (image_t M, image_t U, r3x3_t *map, image_t N, image_t V)
{
    int ix, iy, tx, ty;

    if (U != NULL) { assert(M->nx == U->nx); }
    if (U != NULL) { assert(M->ny == U->ny); }
    
    if (V != NULL) { assert(N->nx == V->nx); }
    if (V != NULL) { assert(N->ny == V->ny); }
    
    /* Compute the map that takes {N} points to {M} points: */
    r3x3_t inv_map;
    r3x3_inv (map, &inv_map);

    /* Choose the number of sampling points along each axis, steps, and the weights: */
    int hw_x, hw_y;
    double *w_x, *w_y;
    double step_x, step_y;
    image_compute_map_weights(inv_map.c[1][1], inv_map.c[2][1], &w_x, &hw_x, &step_x);
    image_compute_map_weights(inv_map.c[1][2], inv_map.c[2][2], &w_y, &hw_y, &step_y);
 
    /* Loop on pixels of output image {N}: */
    for (ix = 0; ix < N->nx; ix++) {
        for (iy = 0; iy < N->ny; iy++) {
	     double px, py;
            image_pixel_center (ix, iy, &px, &py);
          
             double sum_w_M = 0.0;  /* Weighted sum of {M} pixels that go into this {N} pixel. */
            double sum_w = 0.0; /* Sum of weights of those pixels. */
                 
            for (tx = -hw_x; tx <= +hw_x; tx++) {
                for (ty = -hw_y; ty <= +hw_y; ty++) {
                    double dx = px + step_x*tx;
                    double dy = py + step_y*ty;
                    double gx, gy;
                    image_map_coords (dx, dy, &inv_map, &gx, &gy);
                    /* fprintf(stderr, "    p_N = (%8.3f,%8.3f) p_M = (%8.3f,%8.3f)", dx, dy, gx, gy); */
                    if ((gx >= 0.5) && (gx <= M->nx - 0.5) && (gy >= 0.5) && (gy < M->ny - 0.5)) {
		        double M_val, U_val;
                        /* ??? Escolher uma interpolação melhor ??? */
                        image_eval_bilinear (gx, gy, M, &M_val);
                        if (U != NULL) {image_eval_bilinear (gx, gy, U, &U_val);}
                        else {U_val = 1.0;}
                        double w = U_val * w_x[tx+hw_x] * w_y[ty+hw_y];
			/* fprintf(stderr, "  w = %8.6f M_val = %8.6f U_val = %8.6f\n", w, M_val, U_val); */
                        sum_w_M += w * M_val;
                        sum_w += w;
                    }
		    /* fprintf(stderr, "\n"); */
                }
            }
            double N_val = (sum_w_M + 0.5e-200)/(sum_w + 1.0e-200);
	    double V_val = sum_w;
	    /* fprintf(stderr, "  val = %8.3f\n", N_val); */
	    N->val[iy][ix] = N_val;
	    if (V != NULL) { V->val[iy][ix] = V_val; }
        }
    }
    free(w_x);
    free(w_y);
}

#define MAX_KERNEL_SIZE 601
/* Safety limit - bombs out if exceeded. */

#define MIN_GAUSSIAN_WEIGHT 0.05
/* Ignore gaussian weights less than this value (relative to max value). */

void image_compute_gaussian_weights(double sigma, double **wp, int *hwp)  
{
  /* The weight {w[hw + k]} is {exp(-(k/sigma)^2/2)}. */
  /* The half-size {hw} is such that {w[-1]} would be less than {MIN_GAUSSIAN_WEIGHT}. */
  int hw = (int)ceil(sigma * sqrt(-2*log(MIN_GAUSSIAN_WEIGHT)));
  if (DEBUG_ALL) {
    fprintf(stderr, "gaussian sigma = %6.3lf kernel hw = %d", sigma, hw);
  }
  assert(hw >= 1);
  
  /* Allocate the table: */
  int nw = 2*hw + 1;
  assert(nw <= MAX_KERNEL_SIZE);
  double *w = malloc(nw*sizeof(double));

  /* Fill the table: */
  int k;
  for (k = 0 ; k <= hw ; k++) {
    w[hw+k] = w[hw-k] = exp(-k*k / (2*sigma*sigma));
  }

  /* Normalize: */
  double sum = 1.0e-200;
  int i;
  for (i = 0 ; i < nw ; i++) {  sum += w[i]; }
  if (DEBUG_ALL) { fprintf (stderr, "sum = %f", sum); }
  for (i = 0 ; i < nw ; i++) {  w[i] /= sum;}
  if (DEBUG_ALL) { fprintf (stderr, "\n"); }
  (*hwp) = hw;
  (*wp) = w;
}

#define RESAMPLING_SIGMA 0.4
/* Relative width of resampling kernel (from the KLT tracker of S.T.B.) */

void image_compute_map_weights(double dx, double dy, double **wp, int *hwp, double *stepp)
{
  /* Compute the displacement {r} in the source image due to unit step in result image: */
  double r = hypot(dx, dy); /* ??? Should it be the L1 norm instead? ??? */
  /* Compute the standard deviation {sigma} of the kernel: */
  double sigma = RESAMPLING_SIGMA * r;
  /* Compute the table and its half-width: */
  image_compute_gaussian_weights(sigma, wp, hwp);
  /* Compute the step size in the result image: */
  double step = 1.0/r;
  (*stepp) = step;
}

void image_paint_mapped_mark (image_t gab, image_t wmk, image_t img, double alpha, double beta, r3x3_t *map)
{
    int ix, iy, tx, ty;
    int nw = 5;
    int hw = nw/2;  

    assert(nw == 5);
    double w_tb[5] = {.09549149, .65450848, .99999999, .65450853, .09549153};
    double w_tot = 0.0;

    int k;
    for (k = 0; k < nw; k++) {w_tot += w_tb[k];}
    for (k = 0; k < nw; k++) {w_tb[k] /= w_tot;}

    r3x3_t inv_map;
    r3x3_inv (map, &inv_map);
 
    for (ix = 0; ix < img->nx; ix++) {
        for (iy = 0; iy < img->ny; iy++) {
            double val = img->val[iy][ix]; 
            double sum_ww_gab = 0.0;
            double sum_ww = 1.0e-100;
            double px, py;
            image_pixel_center (ix, iy, &px, &py);
                  
            for (tx = -hw; tx <= +hw; tx++) {
                for (ty = -hw; ty <= +hw; ty++) {
                    double dx = px + 0.33*tx;
                    double dy = py + 0.33*ty;
                    double gx, gy;
                    image_map_coords (dx, dy, &inv_map, &gx, &gy);
                    if ((gx >= 0) && (gx < gab->nx) && (gy >= 0) && (gy < gab->ny)) {

                        double w, g;
                        image_eval_bilinear (gx, gy, wmk, &w);
                        image_eval_bilinear (gx, gy, gab, &g);

                        double ww = w * (w_tb[tx+hw] * w_tb[ty+hw]);
                        sum_ww_gab += ww * (alpha*g + beta);
                        sum_ww += ww;
                    }
                }
                val = (1 - sum_ww)*val + sum_ww_gab;  
                img->val[iy][ix] = val;
            }
        }
    }
}

void image_paint_translated_mark (image_t gab, image_t wmk, image_t img, int subsampling, double alpha, double beta, r2_t p)
{
    /* Translation of {img} point to {gab} point: */
    double Gnx = gab->nx;
    double Gny = gab->ny;
    r3x3_t map;
    map.c[0][0] = 1.0; 
    map.c[0][1] = 0.0; 
    map.c[0][2] = 0.0; 
    map.c[1][0] = p.c[0] - Gnx/2.0/subsampling; 
    map.c[1][1] = 1.0/subsampling; 
    map.c[1][2] = 0.0; 
    map.c[2][0] = p.c[1] - Gny/2.0/subsampling; 
    map.c[2][1] = 0.0; 
    map.c[2][2] = 1.0/subsampling;

    image_paint_mapped_mark (gab, wmk, img, alpha, beta, &map);
 
}

void image_extract_subimage(image_t img, image_t gab, int minx, int miny)
{
    int jx, jy;

    int nx = gab->nx;
    int ny = gab->ny;

    for (jx = 0; jx < nx; jx++) {
        for (jy = 0; jy < ny; jy++) {
          /* Nominal indices of source pixel: */
	  int ix = jx + minx;
          int iy = jy + miny;
          /* Clip to domain of {img}: */
          if (ix < 0) { ix = 0; }
	  if (ix >= img->nx) {ix = img->nx-1; }
          if (iy < 0) { iy = 0; }
	  if (iy >= img->ny) {iy = img->ny-1; }
          gab->val[jy][jx] = img->val[iy][ix];
        }
    }
}

image_t image_magnify (image_t img, int magnify)
{
    int x, y;
    
    int bigcols = (img->nx)*magnify;
    int bigrows = (img->ny)*magnify;

    image_t out = image_alloc(bigcols, bigrows);

    for (y = 0; y < bigrows; y++) 
    for (x = 0; x < bigcols; x++) {
      double v, oldx, oldy;
      oldx = ((double)x + 0.5)/((double)magnify);
      oldy = ((double)y + 0.5)/((double)magnify);
      image_eval_bilinear (oldx, oldy, img, &v);
      out->val[y][x] = v;
    }
    return out;
}

image_t image_extract_subimage_by_pos (image_t img, r2_t pi, int radius_x, int radius_y, int magnify)
{
    int jx, jy;
    
    int newrows = (radius_x*2);
    int newcols = (radius_y*2);
    int bigcols = (radius_x*2)*magnify;
    int bigrows = (radius_y*2)*magnify;

    image_t tmp = image_alloc(newcols, newrows);
    image_t out = image_alloc(bigcols, bigrows);

    /*for (y = 0; y < bigrows; y++) 
    for (x = 0; x < bigcols; x++) {
      double v;
      oldx = ((double)x + 0.5)/((double)magnify);
      oldy = ((double)y + 0.5)/((double)magnify);
      image_eval_bilinear (oldx, oldy, img, &v);
      redimg->val[y][x] = v;
      blueimg->val[y][x] = v;
      greenimg->val[y][x] = v;
    }*/


    for (jy = 0; jy < newrows; jy++) {
       for (jx = 0; jx < newcols; jx++) {
          /* Nominal indices of source pixel: */
          double v, oldx, oldy;

          oldx = (double)((pi.c[0] + jx - radius_x) + 0.5);
          oldy = (double)((pi.c[1] + jy - radius_y) + 0.5);

          if ((oldx <= 0) || (oldx >= img->nx))
              assert(1);
          if ((oldy <= 0) || (oldy >= img->ny))
              assert(1);

          image_eval_bilinear (oldx, oldy, img, &v); 

          tmp->val[jy][jx] = v;
        }
    }


    for (jy = 0; jy < bigrows; jy++) {
       for (jx = 0; jx < bigcols; jx++) {
          /* Nominal indices of source pixel: */
          double v, oldx, oldy;

          oldx = ((double)jx + 0.5)/((double)magnify);
          oldy = ((double)jy + 0.5)/((double)magnify);

          //oldx = pi.c[0] + ((double)((jx - radius_x)) + 0.5)/((double)magnify);
          //oldy = pi.c[1] + ((double)((jy - radius_y)) + 0.5)/((double)magnify);

          //printf("pi.c[0] %f pi.c[1] %f oldx %f oldy %f\n", pi.c[0], pi.c[1], oldx, oldy);

          if ((oldx <= 0) || (oldx >= img->nx))
              assert(1);
          if ((oldy <= 0) || (oldy >= img->ny))
              assert(1);

          //image_eval_bilinear (oldx, oldy, img, &v); 
          image_eval_bilinear (oldx, oldy, tmp, &v);

          out->val[jy][jx] = v;
        }
    }
    return out;
}


void image_extract_mapped_mark (image_t img, image_t gab, double alpha, double beta, r3x3_t *map)
{
    int jx, jy;

    int nx = gab->nx;
    int ny = gab->ny;

    for (jx = 0; jx < nx; jx++) {
        for (jy = 0; jy < ny; jy++) {
            double gx, gy;
            image_pixel_center (jx, jy, &gx, &gy);
	    double px, py; 
            image_map_coords (gx, gy, map, &px, &py);
            double v;
            image_eval_bilinear (px, py, img, &v);
            gab->val[jy][jx] = (v - beta)/alpha;
        }
    }
}

void image_extract_translated_mark (image_t img, image_t gab, int subsampling, double alpha, double beta, r2_t p)
{
    int jx, jy;

    int nx = gab->nx;
    int ny = gab->ny;

    for (jx = 0; jx < nx; jx++) {
        for (jy = 0; jy < ny; jy++) {
            double gx, gy;
            image_pixel_center (jx, jy, &gx, &gy);
	    double px, py; 
            image_map_mark_coords_to_image_coords(gx, gy, nx, ny, subsampling, p, &px, &py);
            double v;
            image_eval_bilinear (px, py, img, &v);
            gab->val[jy][jx] = (v - beta)/alpha;
        }
    }
}

void image_apply_color_correction (image_t img, double alpha, double beta)
{
    int jx, jy;

    int nx = img->nx; 
    int ny = img->ny;

    for (jx = 0; jx < nx; jx++) {
        for (jy = 0; jy < ny; jy++) {
            img->val[jy][jx] = (img->val[jy][jx]*alpha) + beta;
        }
    }
}

void image_fill (image_t img, double value)
{
    int iy, ix;

    for (iy = 0; iy < img->ny; iy++) {
        for (ix = 0; ix < img->nx; ix++) {
            img->val[iy][ix] = value;
        }
    }
}

double ** image_mdouble_alloc (int nrows, int ncols)
{
    int row;
    double **m = (double **)notnull(malloc(nrows * sizeof (double *)), "out of mem");
    for (row = 0; row < nrows; row++){
      m[row] = (double *)notnull(malloc(ncols * sizeof(double)), "out of mem");
    }
    return m;
}


void image_mdouble_free (double **m, int nrows) 
{
    int row;
    for (row = 0; row < nrows; row++){ free(m[row]); }
    free(m);
}


image_t image_blur (image_t img, image_t msk, double blur)
{
  image_t res = image_copy(img);
 

  if (blur == 0.0) return res;

  if (msk != NULL) {
    assert(msk->nx == img->nx);
    assert(msk->ny == img->ny);
  }

    int ix, iy, tx, ty;
    
    /* Choose the number of sampling points along each axis, steps, and the weights. */
    int hw;
    double *w_tb;
    image_compute_blur_weights(blur, &hw, &w_tb);
    if (DEBUG_ALL) {
      fprintf(stderr, "image_blur: hw = %d res->nx = %d res->ny %d\n", hw, res->nx, res->ny); 
    }
 
    for (ix = 0; ix < res->nx; ix++) {
        for (iy = 0; iy < res->ny; iy++) {
            double sum_w_img = 0.0;
            double sum_w = 1.0e-100;
            for (tx = -hw; tx <= +hw; tx++) {
                for (ty = -hw; ty <= +hw; ty++) {
                    int dx = ix + tx;
                    int dy = iy + ty;
                    /* fprintf(stderr, "    p = (%4d,%4d)", tx, ty); */
                    if ((dx >= 0) && (dx < img->nx) && (dy >= 0) && (dy < img->ny)) {
		        double img_val = img->val[dy][dx];
                        double msk_val = (msk==NULL?1.0:msk->val[dy][dx]); 
                        double w = w_tb[tx+hw] * w_tb[ty+hw] * msk_val;
			/* fprintf(stderr, "  w = %8.6f val = %8.6f", w, img_val); */
                        sum_w_img += w * img_val;
                        sum_w += w;
                    }
                }
            }
            double res_val = sum_w_img/sum_w;
	    res->val[iy][ix] = res_val;
        }
    }
    free(w_tb);
    return res;
}

void image_compute_blur_weights(double blur, int *hwp, double **wp)
{
  double tol = 0.00001; 
  int hw = (int)floor(blur*sqrt(-2*log(tol)));
  if (hw < 0) { hw = 0; }
  int nw = 2*hw + 1;
  double *w = malloc(nw*sizeof(double));
  int t;
  if (DEBUG_ALL) { fprintf(stderr, "blur weights:"); }
  for (t = -hw; t <= hw; t++){ 
     double z = ((double)t)/blur; w[t+hw] = exp(-z*z/2);
     if (DEBUG_ALL) { fprintf(stderr, "%f", w[t+hw]); }
  }
  if (DEBUG_ALL) { fprintf(stderr, "\n"); }
  (*hwp) = hw;
  (*wp) = w;
}
 
void image_weighted_avg_var (image_t I, image_t W, int subsampling, r2_t p_i, double *avg, double *var)
{
    int jx, jy;
    int Wnx = W->nx;
    int Wny = W->ny;

    double sum_W = 0.0, sum_W_I = 0.0, sum_W_I2 = 0.0;

    for (jx = 0; jx < Wnx; jx++) {
        for (jy = 0; jy < Wny; jy++) {
            double W_val = W->val[jy][jx];
            double gx, gy, px, py; 
            image_pixel_center (jx, jy, &gx, &gy);
            image_map_mark_coords_to_image_coords(gx, gy, Wnx, Wny, subsampling, p_i, &px, &py); 
            double I_val;
            image_eval_bilinear (px, py, I, &I_val);
            sum_W += W_val;
            sum_W_I += I_val * W_val;
            sum_W_I2 += I_val * I_val * W_val;
        }
    }
    *avg =  sum_W_I / sum_W;
    *var = (sum_W_I2 / sum_W) - (*avg)*(*avg);
}

void image_weighted_regression
  ( image_t W, 
    image_t M, 
    int M_subsampling, 
    r2_t M_position, 
    image_t I, 
    int I_subsampling, 
    r2_t I_position,
    bool_t debug, 
    double *alpha, 
    double *beta, 
    double *corr )
{
    int jx, jy;
    int Wnx = W->nx;
    int Wny = W->ny;

    double sum_W = 0.0;
    double sum_W_I = 0.0, sum_W_I2 = 0.0;
    double sum_W_M = 0.0, sum_W_M2 = 0.0;
    double sum_W_M_I = 0.0;

    for (jx = 0; jx < Wnx; jx++) {
        for (jy = 0; jy < Wny; jy++) {
            double W_val = W->val[jy][jx];
            double gx, gy, M_px, M_py, I_px, I_py; 
            image_pixel_center (jx, jy, &gx, &gy);
            image_map_mark_coords_to_image_coords(gx, gy, Wnx, Wny, M_subsampling, M_position, &M_px, &M_py); 
            image_map_mark_coords_to_image_coords(gx, gy, Wnx, Wny, I_subsampling, I_position, &I_px, &I_py); 
            double I_val, M_val;
            image_eval_bilinear (I_px, I_py, I, &I_val);
            image_eval_bilinear (M_px, M_py, M, &M_val);
            sum_W += W_val;
            sum_W_I += I_val * W_val;
            sum_W_I2 += I_val * I_val * W_val;
            sum_W_M += M_val * W_val;
            sum_W_M2 += M_val * M_val * W_val;
            sum_W_M_I += M_val * I_val * W_val;
        }
    }

    double det =       sum_W_M2 * sum_W - sum_W_M * sum_W_M;
    double num_alpha = sum_W_M_I * sum_W - sum_W_I * sum_W_M;
    double num_beta  = sum_W_M2 * sum_W_I - sum_W_M * sum_W_M_I;
    if (det == 0) { det = 1.0e-8; }
    *alpha = num_alpha/det;
    *beta = num_beta/det;

    if (debug) {
      fprintf(stderr, "Sistema alpha-beta:\n");
      fprintf(stderr, "  %18.8f %18.8f   %18.8f\n", sum_W_M2, sum_W_M, sum_W_M_I);
      fprintf(stderr, "  %18.8f %18.8f   %18.8f\n", sum_W_M,  sum_W,   sum_W_I);
      fprintf(stderr, "  alpha = %18.8f beta = %18.8f\n", *alpha, *beta);
    }

    if (sum_W == 0) { sum_W = 1.0e-8; }
    double I_avg =  sum_W_I / sum_W;
    double M_avg =  sum_W_M / sum_W;
    double I_var = (sum_W_I2 / sum_W) - I_avg*I_avg;
    double M_var = (sum_W_M2 / sum_W) - M_avg*M_avg;
    if (I_var == 0) { I_var = 1.0e-8; }
    double corr2 = ((*alpha)*(*alpha)*M_var)/I_var;
    (*corr) = sqrt(fmax(0.0, corr2));
    if (debug) {
      fprintf(stderr, "  I_avg = %18.8f  I_var = %18.8f  I_dev = %18.8f\n", I_avg, I_var, sqrt(fabs(I_var)));
      fprintf(stderr, "  M_avg = %18.8f  M_var = %18.8f  M_dev = %18.8f\n", M_avg, M_var, sqrt(fabs(M_var)));
      fprintf(stderr, "  corr2 = %18.8f corr = %18.8f\n", corr2, *corr);
    }
}

void image_soft_resize (image_t M, image_t U, int redFactor, image_t *N, image_t *V)
{
  /* Compute size {Nnx,Nny} of reduced image: */
  int Nnx = M->nx/redFactor;
  int Nny = M->ny/redFactor;
  
  /* Create the reduced images: */
  *N = image_alloc(Nnx, Nny);
  if (V != NULL)  { *V = image_alloc(Nnx, Nny); }
  
  /* Build an affine map that performs the desired reduction: */
  r3x3_t map = image_resize_map(M->nx, M->ny, redFactor);

  /* Paint images {M,U} into {N,V}: */
  image_map(M, U, &map, *N, (V == NULL ? NULL : *V));
}

r3x3_t image_resize_map(int nx, int ny, int redFactor)
{
  /* Centering shifts (note: integer division!): */ 
  int qx = ((nx%redFactor)+1)/2;
  int qy = ((ny%redFactor)+1)/2;

  r3x3_t map;
  map.c[0][0] = 1.0; 
  map.c[0][1] = 0.0; 
  map.c[0][2] = 0.0; 
  map.c[1][0] = (-qx - 0.5)/redFactor + 0.5; 
  map.c[1][1] = 1.0/redFactor; 
  map.c[1][2] = 0; 
  map.c[2][0] = (-qy - 0.5)/redFactor + 0.5; 
  map.c[2][1] = 0;
  map.c[2][2] = 1.0/redFactor;
  return map;
}

r2_t image_map_point_to_smaller_image (r2_t p_i, int dscale, int redFactor, int nx, int ny)
{
  assert(dscale >= 0);
  if (dscale > 0)
    { /* Centering shifts: */ 
      int qx = ((nx%redFactor)+1)/2;
      int qy = ((ny%redFactor)+1)/2;

      /* Map {p_i} from the large image to the reduced image */
      p_i.c[0] = (p_i.c[0] - qx - 0.5)/redFactor + 0.5;
      p_i.c[1] = (p_i.c[1] - qy - 0.5)/redFactor + 0.5;

      return image_map_point_to_smaller_image
        ( p_i, dscale-1, redFactor, nx/redFactor, ny/redFactor );
    }
  else
    { return p_i; }
}

r2_t image_map_point_to_larger_image (r2_t p_r, int dscale, int redFactor, int nx, int ny)
{
  assert(dscale >= 0);
  if (dscale > 0)
    { 
      p_r = image_map_point_to_larger_image
        ( p_r, dscale-1, redFactor, nx/redFactor, ny/redFactor );

      /* Centering shifts: */ 
      int qx = ((nx%redFactor)+1)/2;
      int qy = ((ny%redFactor)+1)/2;

      /* Map {p_i} from the large image to the reduced image */
      p_r.c[0] = (p_r.c[0] - 0.5)*redFactor + qx + 0.5;
      p_r.c[1] = (p_r.c[1] - 0.5)*redFactor + qy + 0.5;
      return p_r;
    }
  else
    { return p_r; }
}

void image_deform(image_t M, image_t U, r2_t *u_i, r2_t *v_i, int subsampling, image_t *N, image_t *V)
{
  /* Compute size {Nnx,Nny} of deformed image: */
  int Nnx = (int)ceil(2.0*(fabs(u_i->c[0]) + fabs(v_i->c[0]))) * subsampling;
  int Nny = (int)ceil(2.0*(fabs(u_i->c[1]) + fabs(v_i->c[1]))) * subsampling;
  /* Create the deformed images: */
  *N = image_alloc(Nnx, Nny);
  if (V != NULL)  { *V = image_alloc(Nnx, Nny); }
  /* Build an affine map that performs the desired domain deformation: */
  r3x3_t map = image_map_from_shape_vectors(M->nx, M->ny, Nnx, Nny, u_i, v_i, subsampling);
  /* Paint images {M,U} into {N,V}: */
  image_map(M, U, &map, *N, (V == NULL ? NULL : *V));
}

r3x3_t image_map_from_shape_vectors
  ( int Mnx, 
    int Mny,
    int Gnx, 
    int Gny, 
    r2_t *u_i, 
    r2_t *v_i,
    int subsampling )
{
     r3x3_t map;
     r2_t us_i, vs_i;
     r2_scale(+2.0/Mnx, u_i, &us_i);
     r2_scale(-2.0/Mny, v_i, &vs_i);
     map.c[0][0] = 1.0;
     map.c[0][1] = 0.0;
     map.c[0][2] = 0.0;
     map.c[1][0] = Gnx/2.0 - subsampling*u_i->c[0] + subsampling*v_i->c[0];
     map.c[1][1] = subsampling*us_i.c[0];
     map.c[1][2] = subsampling*vs_i.c[0];
     map.c[2][0] = Gny/2.0 - subsampling*u_i->c[1] + subsampling*v_i->c[1];
     map.c[2][1] = subsampling*us_i.c[1];
     map.c[2][2] = subsampling*vs_i.c[1];
     if (DEBUG_ALL) {
       fprintf(stderr, "%s: map =\n", __FUNCTION__);
       r3x3_print(stderr, &map);
     }
     return map;
}  


void image_subtract_mark (image_t I, image_t M)
{
    int jx, jy;
    int Inx = I->nx;
    int Iny = I->ny;
    int Mnx = M->nx;
    int Mny = M->ny;

    assert(((Inx-Mnx) % 2) == 0);
    assert(((Iny-Mny) % 2) == 0);

    int dx = (Inx-Mnx)/2; 
    int dy = (Iny-Mny)/2;
 
    for (jx = 0; jx < Inx; jx++) {
        for (jy = 0; jy < Iny; jy++) {
            int gx = jx-dx;
            int gy = jy-dy;
            if ((gx >= 0) && (gx < Mnx) && (gy >= 0) && (gy < Mny))
                I->val[jy][jx] -= M->val[gy][gx];
        }
    }  
}

void image_multiply_by_normalized_weight (image_t I, image_t W)
{
  int wx, wy;

  /* Find max weight: */
  int Wnx = W->nx;
  int Wny = W->ny;

  double wmax = -HUGE_VAL;
  for (wy = 0; wy < Wny; wy++) {
    for (wx = 0; wx < Wnx; wx++) {
      wmax = fmax(wmax, W->val[wy][wx]);
    }
  }
  assert(wmax > 0.0);

  /* Multiply {W} normalized into {I}: */
  int Inx = I->nx; assert(((Inx-Wnx) % 2) == 0);
  int Iny = I->ny; assert(((Iny-Wny) % 2) == 0);

  int dx = (Inx-Wnx)/2; 
  int dy = (Iny-Wny)/2;
  int ix, iy;
 
  for (ix = 0; ix < Inx; ix++) {
    for (iy = 0; iy < Iny; iy++) {
      /* Locate pixel {wx,wy} of {W} corresponding to pixel {ix,iy} of {I}: */
      int wx = ix-dx;
      int wy = iy-dy;
      if ((wx >= 0) && (wx < Wnx) && (wy >= 0) && (wy < Wny))
	I->val[iy][ix] *= (W->val[wy][wx]/wmax);
      else
	I->val[iy][ix] = 0.0;
    }
  }  
}


r2_t image_extrapolate_position (r2_t p_2, r2_t p_1)
{
  r2_t p;
  p.c[0] = 2*p_1.c[0] - p_2.c[0];
  p.c[1] = 2*p_1.c[1] - p_2.c[1];
  return p;
}

r2_t image_snd_order_extrapolate_position (r2_t p_3, r2_t p_2, r2_t p_1)
{
  r2_t p;
  p.c[0] = p_1.c[0] + 0.5 * (p_1.c[0] - p_3.c[0]) - 3 * (p_2.c[0] - 0.5 * (p_1.c[0] + p_3.c[0]));
  p.c[1] = p_1.c[1] + 0.5 * (p_1.c[1] - p_3.c[1]) - 3 * (p_2.c[1] - 0.5 * (p_1.c[1] + p_3.c[1]));
  return p;
}

