#include "image_STB.h"
#include "image.h"
#include "numeric.h"

void image_eval_STB_old
  ( double x, 
    double y, 
    image_t V, 
    double *v, 
    double *dx_v, 
    double *dy_v,
    image_t W,
    double *w
  )
{
  /* Adjust (x,y) for the (0.5,0.5) pixel center shift: */
  /* ??????? Check */
  x -= 0.5;
  y -= 0.5;

  int ix = floor(x);
  int iy = floor(y);
  double xfrac = x - ix;
  double yfrac = y - iy;

  int t_ok = (v != NULL);
  int dx_t_ok = (dx_v != NULL);
  int dy_t_ok = (dy_v != NULL);

  /* Bilinear interpolation of image values: */
  double t[4], dx_t[4], dy_t[4]; /* Values and derivatives at four surrounding pixels. */
  int jx, jy;
  for (jx = 0; jx < 2; jx++)
    for (jy = 0; jy < 2; jy++) {
      image_eval_aux_STB_old
        ( V, ix+jx, iy+jy, 
          (t_ok ? &(t[jx + 2*jy]) : NULL),
          (dx_t_ok ? &(dx_t[jx + 2*jy]) : NULL),
          (dy_t_ok ? &(dy_t[jx + 2*jy]) : NULL)
	);
    }
  if (v != NULL) {
    (*v) = 
       (1 - xfrac) * (1 - yfrac) * t[0] +
            xfrac  * (1 - yfrac) * t[1] +
       (1 - xfrac) *      yfrac  * t[2] +
            xfrac  *      yfrac  * t[3];
  }
  if (dx_v != NULL) {
    (*dx_v) = 
       (1 - xfrac) * (1 - yfrac) * dx_t[0] +
            xfrac  * (1 - yfrac) * dx_t[1] +
       (1 - xfrac) *      yfrac  * dx_t[2] +
            xfrac  *      yfrac  * dx_t[3];
  }
  if (dy_v != NULL) {
    (*dy_v) = 
       (1 - xfrac) * (1 - yfrac) * dy_t[0] +
            xfrac  * (1 - yfrac) * dy_t[1] +
       (1 - xfrac) *      yfrac  * dy_t[2] +
            xfrac  *      yfrac  * dy_t[3];
  }
  
  (*w) = 1.0;
}

void image_eval_STB_new
  ( double x, 
    double y, 
    image_t V, 
    double *v, 
    double *dx_v, 
    double *dy_v,
    image_t W,
    double *w
  )
{
  /* This procedure performs a biquadratic interpolation of 
    pixel values of an implicitly smoothed image, then differentiates
    the interpolation. */

  /* What do we need to compute: */
  int v_ok = (v != NULL);
  int dx_v_ok = (dx_v != NULL);
  int dy_v_ok = (dy_v != NULL);
  int w_ok = (w != NULL);
  if (! (v_ok | dx_v_ok | dy_v_ok | w_ok)) { return; }

  /* Compute the indices {ix,iy} of pixel that contains {x,y}: */
  int ix = floor(x);
  int iy = floor(y);

  /* Compute local coordinates {xfrac,yfrac} relative to the pixel. */
  double xfrac = x - ix;
  double yfrac = y - iy;
  //fprintf(stderr, "image_eval_STB_new fracs = (%8.6f %8.6f)\n", xfrac, yfrac);

  /* Smoothing implicitly extends the image {V} and the weights {W}
    by a few pixels all around their original domains; the B-spline
    interpolation shrinks them again by about one pixel. */

  /* Compute values and weights of 3x3 pixels  centered at {ix,iy}, with smoothing: */
  double vtxy[9]; /* Pixel values. */
  double wtxy[9]; /* Reciprocals of pixel weights. */
  int jx, jy;
  for (jx = -1; jx <= +1; jx++) {
    for (jy = -1; jy <= +1; jy++) {
      int ind = jy*3 + jx + 4;
      double vt, wt;
      image_eval_aux_STB_new (V, W, ix+jx, iy+jy, &vt, &wt);
      vtxy[ind] = vt;
      wtxy[ind] = (wt == 0 ? +HUGE_VAL : 1/wt);
      //fprintf(stderr, "[%5.3f %5.3f --> %8.1f]", vtxy[ind], wt, wtxy[ind]);
    }
    //fprintf(stderr, "\n");
  }

  /* Interpolate vertically with biquadratic B-spline along columns {ix-1,ix,ix+1}: */
  double vtx[3];
  double dy_vtx[3];
  double wtx[3];
  for (jx = -1; jx <= +1; jx++) {
    double v_tmp[3];
    v_tmp[0] = vtxy[jx+1]; 
    v_tmp[1] = vtxy[jx+4]; 
    v_tmp[2] = vtxy[jx+7]; 
    interpol_spline_2_B
      ( yfrac, 
        v_tmp, 
        &(vtx[jx+1]), 
        &(dy_vtx[jx+1]),
        NULL
      );
    double w_tmp[3];
    w_tmp[0] = wtxy[jx+1]; 
    w_tmp[1] = wtxy[jx+4]; 
    w_tmp[2] = wtxy[jx+7]; 

    interpol_spline_2_B
      ( yfrac, 
        w_tmp, 
        &(wtx[jx+1]), 
        NULL,
        NULL
      );
  }

  /* Interpolate horizontally: */
  if ((v_ok) || (dx_v_ok)) {
    interpol_spline_2_B(xfrac, vtx, v, dx_v, NULL);
    if (isnan(*v) || (fabs(*v) > 1.0e20)) { image_STB_debug(*v); }
  }
  
  if (dy_v_ok) {
    interpol_spline_2_B(xfrac, dy_vtx, dy_v, NULL, NULL);
  }

  /* Compute weight {*w} of result. For now, return the minimum of all 9 weights. */
  if (w_ok) {
    double wtxy;
    interpol_spline_2_B(xfrac, wtx, &wtxy, NULL, NULL);
    (*w) = (wtxy == +HUGE_VAL ? 0.0 : 1/wtxy);
  }
}

void image_eval_aux_STB_old
  ( image_t V, 
    int ix, 
    int iy, 
    double *v,
    double *dx_v,
    double *dy_v
  )
{
  int nx = V->nx;
  int ny = V->ny;

  /* The image itself is not smoothed: */
  double t = ((ix >= 0) && (ix < nx) && (iy >= 0) && (iy < ny) ? V->val[iy][ix] : 0 );

  /* !!! Should skip some computations if {v,dx_v,dy_v} are NULL. */

  int jx, jy;

  /* Half-width and width of weight mask for Gaussian-smoothed derivative. */
  #define HNGW 3
  #define NGW (2*HNGW + 1)

  /* Gaussian smoothing weights (normalized to unit sum): */
  double gw[NGW] = {
    0.004433,
    0.054006,
    0.242036,
    0.399050,
    0.242036,
    0.054006,
    0.004433
  };

  /* Derivatives of smoothing weights: */
  double dgw[NGW] = {
    -0.013354,
    -0.108455,
    -0.243030,
    00.000000,
    +0.243030,
    +0.108455,
    +0.013354
  };
  double dx_t = 0, dy_t = 0;
  for(jx = -HNGW; jx <= +HNGW; jx++){
    /* Vertical smoothing in column {ix+jx}, result to {tx,dy_tx}: */
    double tx = 0.0;
    double dy_tx = 0.0;
    if ((ix+jx >= 0) && (ix+jx < nx)) {
      for(jy = -HNGW; jy <= +HNGW; jy++){
        if ((iy+jy >= 0) && (iy+jy < ny)) {
          int indy = jy + HNGW;
          tx += V->val[iy+jy][ix+jx]*gw[indy];
          dy_tx += V->val[iy+jy][ix+jx]*dgw[indy];
	}
      }
      /* Accumulate over columns: */
      int indx = jx + HNGW;
      dx_t += tx*dgw[indx];
      dy_t += dy_tx*gw[indx];
    }
  }

  if (v != NULL) { (*v) = t; }
  if (dx_v != NULL) { (*dx_v) = dx_t; }
  if (dy_v != NULL) { (*dy_v) = dy_t; }
  #undef HNGW
  #undef NGW
}

void image_eval_aux_STB_new
  ( image_t V, 
    image_t W, 
    int ix, 
    int iy, 
    double *v,
    double *w
  )
{
  int nx = V->nx; if (W != NULL) { assert(nx == W->nx); }
  int ny = V->ny; if (W != NULL) { assert(ny == W->ny); }

  int jx, jy;

  /* Half-width and width of weight mask for Gaussian-smoothed derivative. */
  #define HNGW 3
  #define NGW (2*HNGW + 1)

  /* Gaussian smoothing weights: */
  double gw[NGW] = {
    0.004433,
    0.054006,
    0.242036,
    0.399050,
    0.242036,
    0.054006,
    0.004433
  };
  double sum_wv = 0.0;
  double sum_w = 0.0;
  for(jx = -HNGW; jx <= +HNGW; jx++){
    if ((ix+jx >= 0) && (ix+jx < nx)) {
      int indx = jx + HNGW;
      for(jy = -HNGW; jy <= +HNGW; jy++){
        if ((iy+jy >= 0) && (iy+jy < ny)) {
          int indy = jy + HNGW;
          double wxy = (W == NULL ? 1.0 : W->val[iy+jy][ix+jx]) * gw[indx] * gw[indy];
          double vxy = V->val[iy+jy][ix+jx];
          sum_wv += wxy*vxy;
          sum_w += wxy;
            
	}
      }
    }
  }
  
  if (v != NULL) { (*v) = (sum_wv + 0.5e-200)/(sum_w + 1.0e-200); }
  if (w != NULL) { (*w) = sum_w; }
  #undef HNGW
  #undef NGW
}


void image_STB_debug(double x) {
  assert(! isnan(x));
  return;
}
