#ifndef __image_STB_H__
#define __image_STB_H__

/* Image evaluation functions for the Stanley T. Birchfield's LK goal function. */

#include <image.h>

void image_eval_STB_old
  ( double x, 
    double y, 
    image_t V, 
    double *v, 
    double *dx_v, 
    double *dy_v,
    image_t W,
    double *w
  );
  /* Evaluates the image (target or mark) {V} at the fractional point {x,y}.  Assumes
    that pixel centers are at half-integer coordinates.  The result is
    C0-continuous (a biquadratic B-spline).
    Returns the result in {*v}.  Also returns in {*dx_v,*dy_v}
    the spatial derivatives of the value. 

    This procedure ignores {W} and uses bilinear interpolation of
    pixel values of an implicitly smoothed image and pixel values of
    an implicitly smoothed derivative image.  It returns {*w = 1}. */

void image_eval_STB_new
  ( double x, 
    double y, 
    image_t V, 
    double *v, 
    double *dx_v, 
    double *dy_v,
    image_t W,
    double *w
  );
  /* Evaluates the image (target or mark) {V} at the fractional point {x,y}.  Assumes
    that pixel centers are at half-integer coordinates.  The result is
    C0-continuous (a biquadratic B-spline).
    Returns the result in {*v}.  Also returns in {*dx_v,*dy_v}
    the spatial derivatives of the value. 

    If {W} is not NULL, it must be an image with the same size as {V},
    containing a non-negative weight (``confidence'', ``importance'',
    etc.)  for each pixel.  Pixels outside the domain of {V} and {W}
    are assumed to have weight 0.  If {W} is NULL, the weights are
    assumed to be 1 inside the domain, 0 outside.
    
    The weight of each pixel is taken into account while smoothing 
    the image.

    The procedure returns in {*w} the weight of the result {*v}, defined
    as the the minimum of the weights of all pixels of {V}
    that were used in the computation of {*v}.  */

void image_eval_aux_STB_old
  ( image_t V, 
    int ix, 
    int iy, 
    double *v,
    double *dx_v,
    double *dy_v
  );
  /* Evaluates the image {V} and its gradient images 
    at the pixel {[iy][ix]} with Gaussian smoothing. */

void image_eval_aux_STB_new
  ( image_t V, 
    image_t W, 
    int ix, 
    int iy, 
    double *v,
    double *w
  );
  /* Evaluates the image {V} at the pixel {[iy][ix]} with Gaussian smoothing.
    If the mask {W} is not null, each pixel of {V} is also weighted with the 
    corresponding pixel of {W} (besides the smoothing kernel weight).
    If {W} is NULL, assumes all its elements are 1.  Pixels outside the domain
    of {V} and {W} are assumed to have zero weight.  In any case, the procedure
    also returns in {*w} the sum of all pixel weights (kernel times mask) 
    that entered into the result {*v}.  In particular, if the whole kernel
    falls outside the mask and/or the domain of {V}, the result {*v}
    will be indeterminate, and {*w} will be 0. */    

void image_STB_debug(double x);
/* A bogus procedure for {gdb} breakpoints. */

#endif
