#define _GNU_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include "image_list.h"

image_list_t image_list_alloc (int nimages)
{
    image_list_t ilist;
    ilist = (image_list_t)malloc(sizeof(struct _image_list_t));
    ilist->nimages = nimages;
    ilist->imgs = (image_t *)malloc(nimages * sizeof(image_t));
    int i;
    for (i = 0; i < nimages; i++) { ilist->imgs[i] = NULL; }
    return ilist;
}

void image_list_free (image_list_t ilist)
{
    int i;
    for (i = 0; i < ilist->nimages; i++) { image_free (ilist->imgs[i]); } 
    free(ilist->imgs);
    free(ilist);
}


void image_list_write (image_list_t ilist, int magic, int maxval, char *iname, char *type, double min, double max)
{
    int i;
    for (i = 1; i < ilist->nimages; i++) {
        image_t tmp = image_copy (ilist->imgs[i]);
        if (min != max) image_change_interval (tmp, 0, 0, min, max);
        char *fname = NULL;
        asprintf(&fname,"%s_%03d.%s", iname, i, type);
        image_write_pgm (tmp, magic, maxval, fname);
        free(fname);
        image_free (tmp); 
    }
}

image_list_t image_list_copy_selected(image_list_t T_list, int nsel, int sel[])
{
  image_list_t M_list = image_list_alloc(nsel);
  int k;
  for (k = 0; k < nsel; k++) {
    M_list->imgs[k] = image_copy(T_list->imgs[sel[k]]);
  }
  return M_list;
}

image_list_t image_list_read(int n, char *fname[])
{
  int i;
  int nx = -1, ny = -1;

  image_list_t ilist;

  ilist = (image_list_t)malloc(sizeof(struct _image_list_t));

  ilist->nimages = n;

  ilist->imgs = (image_t *)malloc(n * sizeof(image_t));

  for (i = 0; i < n; i++)
    {
      ilist->imgs[i] = image_read_pgm (fname[i]);

      if (i == 0) { 
	nx = ilist->imgs[i]->nx;
	ny = ilist->imgs[i]->ny;
      }
      else {
	assert(nx == ilist->imgs[i]->nx);
	assert(ny == ilist->imgs[i]->ny);
      }
    }

  return ilist;
}

image_list_t image_list_create_weight_masks (image_list_t M_list, int use_gaussian_weights)
{
  int nmasks = M_list->nimages;
  image_list_t W_list = image_list_alloc(nmasks);
  int k;
  for (k = 0; k < nmasks; k++) {
    image_t M = M_list->imgs[k];
    if (use_gaussian_weights) {
      W_list->imgs[k] = image_gaussian_mask (M->nx, M->ny, 0);
    } else {
      W_list->imgs[k] = image_alloc(M->nx, M->ny);
      image_fill(W_list->imgs[k], 1.0);
    }
  }
  return W_list;
}
