#ifndef __image_list_H__
#define __image_list_H__

#define _GNU_SOURCE
#include <stdio.h>
#include "image.h" 

typedef struct _image_list_t {
  int nimages;   /* Number of images in image list*/
  image_t *imgs; /* Image data */
} *image_list_t;
/* Struct the list of one image */


/* --------------    list functions    --------------*/

image_list_t image_list_alloc (int nimages);
/* Returns a new list with space for {nimages} images. */

void image_list_free (image_list_t ilist);
/* Disallocates a list os images {ilist} */

void image_list_write (image_list_t ilist, int magic, int maxval, char *iname, char *type, double min, double max);

image_list_t image_list_read (int n, char *fname[]);
  /* Reads {n} images from files {fname[0..n-1]}, and returns them as
    an {image_list_t}. */

image_list_t image_list_copy_selected(image_list_t T_list, int nsel, int sel[]);
/* Creates a list of images {M_list} such that {M_list->imgs[k]} is a copy of {T_list->imgs[sel[k]]}
   for {k} in {0..nsel-1}. */

image_list_t image_list_create_weight_masks (image_list_t M_list, int use_gaussian_weights);
/* Creates a list {W_list} of weight masks for the images in {M_list}.  Mask {W_list->imgs[k]}
  has the same size as {M_list->imgs[k]}, for all {k} in {0..M_list->nimages-1}. 
  If {} is true, the masks is a rectangular Gaussian distribution clipped to the 
  image domain.  Otherwise it is a mask of all 1's. */

#endif
