#define _GNU_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include "image_pyramid.h"

image_pyramid_t image_pyramid_alloc_header (int redFactor, int nlevels)
{
    image_pyramid_t pyr;
    pyr = (image_pyramid_t)malloc(sizeof(struct _image_pyramid_t));
    pyr->redFactor = redFactor;
    pyr->nlevels = nlevels;
    pyr->imgs = (image_t *)malloc(nlevels * sizeof(image_t));
    return pyr;
}

void image_pyramid_from_image 
  ( image_t M, 
    image_t U, 
    int redFactor,
     int minSize, 
    image_pyramid_t *M_pyrP, 
    image_pyramid_t *U_pyrP
  )
{
    /* Find the number of levels {nlevels}: */
    int nlevels = 0;
    int ty = M->ny; int tx = M->nx;
    while ((ty >= minSize) && (tx >= minSize)) {
            nlevels++;
            ty /= redFactor; tx /= redFactor;
        }
    if (nlevels == 0) { 
      fprintf(stderr, "initial image is too small (%d x %d)\n", M->nx, M->ny);
      assert(0);
    }

    /* Allocate the output pyramids: */
    image_pyramid_t M_pyr = image_pyramid_alloc_header(redFactor, nlevels);
    image_pyramid_t U_pyr = image_pyramid_alloc_header(redFactor, nlevels);

    /*the first image of the pyramid is the original (input) image*/
    M_pyr->imgs[0] = image_copy(M);
    U_pyr->imgs[0] = (U == NULL ? NULL : image_copy(U));

    int i;

    for (i = 1; i < nlevels; i++) {
        image_t Mprev = M_pyr->imgs[i-1];
        image_t Uprev = U_pyr->imgs[i-1];
        image_t *MthisP = &(M_pyr->imgs[i]);
        image_t *UthisP = &(U_pyr->imgs[i]);
        image_soft_resize(Mprev, Uprev, redFactor, MthisP, UthisP);
    }


    (*M_pyrP) = M_pyr;

    if ((U != NULL) ? ((*U_pyrP) = U_pyr): NULL);
}

void image_pyramid_free (image_pyramid_t pyr)
{
    int i;
    for (i = 0; i < pyr->nlevels; i++) {
        image_free (pyr->imgs[i]);
    } 
    free(pyr->imgs);
    free(pyr);
}

image_pyramid_t image_pyramid_copy (image_pyramid_t I_pyr)
{
    int i;
    image_pyramid_t pyr = image_pyramid_alloc_header(I_pyr->redFactor, I_pyr->nlevels);
    for (i = 0; i < pyr->nlevels; i++) {
        pyr->imgs[i] = image_copy(I_pyr->imgs[i]);
    }
    return pyr;
}

void image_pyramid_write
  ( image_pyramid_t pyr, 
    int magic, 
    int maxval, 
    char *iname, 
    char *tag, 
    char *type, 
    double vmin,
    double vmax, 
    double min, 
    double max
  )
{
    assert(strcmp(type, "pgm") == 0);
    int i;
    for (i = 0; i < pyr->nlevels; i++) {
        image_t tmp = image_copy (pyr->imgs[i]);
        if (min != max) { image_change_interval (tmp, vmin, vmax, min, max); }
        char *fname = NULL;
        asprintf(&fname,"%s_%02d_%s.%s", iname, i, tag, type);
        image_write_pgm (tmp, magic, maxval, fname);
        free(fname);
        image_free (tmp); 
    }
}


