#ifndef __image_pyramid_H__
#define __image_pyramid_H__

#define _GNU_SOURCE
#include <stdio.h>
#include "image.h" 

typedef struct _image_pyramid_t {
  int nlevels;   /* Number of levels in image pyramid*/
  int redFactor;    /* Number of subsampling of the image*/
  image_t *imgs; /* Image data */
} *image_pyramid_t;
/* Struct the pyramid of one image */


/* --------------    pyramid functions    --------------*/
image_pyramid_t image_pyramid_alloc_header (int redFactor, int nlevels);
/* Returns an image pyramid with space for {nlevels} images and the
  specified {redFactor}.  The images are NOT allocated. */

void image_pyramid_from_image 
  ( image_t M, 
    image_t U, 
    int redFactor,
     int minSize, 
    image_pyramid_t *M_pyrP, 
    image_pyramid_t *U_pyrP
  );
/* Returns a pyramid of an image {M}. Level 0 is a copy of {M} (the
   original image). Level 1 is a smaller (coarser) version of {M} with
   {M->nx/redFactor} columns and {M->ny/redFactor} rows, and so on.
   The last level is the smallest version of {M} that has at least
   {minSize} pixels in each direction.

   The images are computed by {image_soft_reduce}. The weight mask
   image {U} is given to {image_soft_reduce}; if {U} is NULL, the
   procedure assumes a mask with all pixels weights equal to 1.  In
   any case, the reduced mask {V} returned by {image_soft_reduce} is
   used for the next step, and all the reduced masks are stored in
   {*U_pyrP}.

  The pyramids {*M_pyrP} and {*U_pyrP} are allocated by the procedure. */

image_pyramid_t image_pyramid_copy (image_pyramid_t I_pyr);
/* Makes a copy of {I_pyr}, copying the images at all levels. */

void image_pyramid_free (image_pyramid_t pyr);
/* Disallocates a list os images {pyr} */

void image_pyramid_write
  ( image_pyramid_t pyr, 
    int magic, 
    int maxval, 
    char *iname, 
    char *tag, 
    char *type, 
    double vmin,
    double vmax, 
    double min, 
    double max
  );
/* Writes out each image in the pyramid {pyr} as a separate file
  of the given {type} (which currently must be "pgm").  The file name
  will be "{iname}_{NN}_{tag}.{type}", where {NN} is the image's level
  in the pyramid (2 digits).  The parameters {vmin,vmax,min,max}
  define the value rescaling, as in {image_change_interval_and_write} */

#endif
