#define _GNU_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <math.h>
#include <image.h>
#include <image_pyramid.h>
#include <tf_calib.h>
#include <r2x2.h>
#include <r2.h>
#include "lk.h" 
#include "lk_error_M.h" 
#include "lk_error_STB.h" 

r3x3_t ti_make_map(double m10, double m11, double m12, double m20, double m21, double m22);
/* Creates an affine map matrix given the elements of the two bottom rows. */

int main (int argc, char *argv[])
{
  /*Verify memory*/
  void *trash = malloc(1);
  struct mallinfo info;
  int MemDinInicial, MemDinFinal;
  /*End*/

  free(trash);
  info = mallinfo();
  MemDinInicial = info.uordblks;

  image_t M = image_read_pgm (argv[1]);
  image_t U;
  if (strcmp(argv[2], "1") == 0)
    { U = NULL; }
  else if (strcmp(argv[2], "G") == 0)
    { U = image_gaussian_mask(M->nx, M->ny, 1); }
  else
    { U = image_read_pgm (argv[2]); }

  double sqhf = sqrt(0.5);

  /* Tests of {image_map} */
  
  int MAXMAPS = 100;
  int redFactor = 2;

  r3x3_t map[MAXMAPS];
  char *nam[MAXMAPS];
  int n = 0;
  nam[n] = "ident"; r3x3_ident(&(map[n++]));
  nam[n] = "redHf"; map[n++] = ti_make_map(00.00, +0.50, 00.00,  00.00, 00.00, +0.50);
  nam[n] = "magTw"; map[n++] = ti_make_map(00.00, +2.00, 00.00,  00.00, 00.00, +2.00);
  nam[n] = "rot90"; map[n++] = ti_make_map(+40.0, 00.00, +1.00,  +280., -1.00, 00.00);
  nam[n] = "redQt"; map[n++] = ti_make_map(00.00, +0.25, 00.00,  00.00, 00.00, +0.25);
  nam[n] = "trans"; map[n++] = ti_make_map(-5.00, +1.00, 00.00,  +9.00, 00.00, +1.00);
  nam[n] = "redNU"; map[n++] = ti_make_map(00.00, +0.50, 00.00,  00.00, 00.00, +0.25);
  nam[n] = "rotNU"; map[n++] = ti_make_map(+130., 00.00, +0.50,  +200., -0.25, 00.00);
  nam[n] = "rot45"; map[n++] = ti_make_map(00.00, +sqhf, +sqhf,  +120., -sqhf, +sqhf);
  nam[n] = "smRed"; map[n++] = image_resize_map(M->nx, M->ny, redFactor);
  assert(n <= MAXMAPS);

  image_t N = image_alloc(M->nx, M->ny);
  image_t V = image_alloc(M->nx, M->ny);
  
  int i;
  for (i = 0; i < n; i++) {
    image_map(M, U, &(map[i]), N, V);

    char *fname_N = NULL;
    asprintf(&fname_N, "out/N_%02d_%s.pgm", i, nam[i]);
    image_write_pgm (N, PGM_RAW_MAGIC, MAXVAL, fname_N);
    free(fname_N);

    char *fname_V = NULL;
    asprintf(&fname_V, "out/V_%02d_%s", i, nam[i]);
    image_change_interval_and_write (V, PGM_RAW_MAGIC, MAXVAL, fname_V, "pgm", 0, 0, 0, 1);
    free(fname_V);
  }

  /* Test image pyramids: */
  int minSize = 3;
  image_pyramid_t M_pyr;
  image_pyramid_t U_pyr;
  image_pyramid_from_image (M, U, redFactor, minSize, &M_pyr, &U_pyr);

  image_pyramid_write(M_pyr, PGM_RAW_MAGIC, MAXVAL, "out/pyr", "M", "pgm", 0,0, 0,0);
  //image_pyramid_write(U_pyr, PGM_RAW_MAGIC, MAXVAL, "out/pyr", "U", "pgm", 0,0, 0,1);
  
  image_free (M);
  if (U != NULL) { image_free (U); }
  image_free (N);
  if (V != NULL) { image_free (V); }
  image_pyramid_free(M_pyr);
  //image_pyramid_free(U_pyr);

  info = mallinfo();
  MemDinFinal = info.uordblks;
  if (MemDinInicial!=MemDinFinal)
    { fprintf(stderr, "** main: memory leak - %d blocks\n", MemDinFinal-MemDinInicial); }
  return 0;
} 

r3x3_t ti_make_map(double m10, double m11, double m12, double m20, double m21, double m22)
{
  r3x3_t map;
  map.c[0][0] = 1.0;
  map.c[0][1] = 0.0;
  map.c[0][2] = 0.0;
  map.c[1][0] = m10;
  map.c[1][1] = m11;
  map.c[1][2] = m12;
  map.c[2][0] = m20;
  map.c[2][1] = m21;
  map.c[2][2] = m22;
  return map;
}
