#define _GNU_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <math.h>
#include <image.h>
#include <lk_error_STB.h>

void plot_2D_image_and_derivatives (char *name, char *method, image_t V, image_t W);

void interpol_2D
  ( char* method, 
    double x, 
    double y, 
    image_t V, 
    double *v, 
    double *dx_v, 
    double *dy_v, 
    double *dxx_v, 
    double *dxy_v, 
    double *dyy_v,
    image_t W, 
    double *w
  );
  /* Given an image {M} and a weight mask {W} (possibly NULL), 
    interpolates them at point {(x,y)} of {M}'s domain with the
    given {method}. */

int main (int argc, char *argv[])
{
  struct mallinfo info = mallinfo();
  int MemDinInicial = info.uordblks;

  char *method = argv[1];

  image_t V = image_read_pgm (argv[2]);
  image_t W = image_read_pgm (argv[3]);

  plot_2D_image_and_derivatives ("out/imeval", method, V, W);

  image_free(V);
  image_free(W);

  info = mallinfo();
  int MemDinFinal = info.uordblks;
  if (MemDinInicial!=MemDinFinal)
    fprintf(stderr, "** main: memory leak - %d blocks\n", MemDinFinal-MemDinInicial);
  return 0;
} 

void plot_2D_image_and_derivatives (char *name, char *method, image_t V, image_t W)
{
  int rx, ry;
  if (W != NULL) { assert(V->nx == W->nx); assert(V->ny == W->ny); }

  fprintf(stderr, "plotting 2D image with method \"%s\"\n", method);

  char *fname = NULL;
  asprintf(&fname, "%s_%s.txt", name, method);
  fprintf(stderr, "writing %s ...", fname); 
  FILE *fplot = fopen(fname, "w");
  assert(fplot != NULL);

  int Nx = 10*V->nx; /* Number of plot intervals in X. */
  int Ny = 10*V->ny; /* Number of plot intervals in Y. */

  for (rx = 0; rx <= Nx; rx++) {
    for (ry = 0; ry <= Ny; ry++) {
      double x = ((double)rx/Nx)*V->nx;
      double y = ((double)ry/Ny)*V->ny;

      double v, dx_v, dy_v, dxx_v, dxy_v, dyy_v, w;

      interpol_2D (method, x, y, V, &v, &dx_v, &dy_v, &dxx_v, &dxy_v, &dyy_v, W, &w); 

      fprintf(fplot, "%12.8f %12.8f  %12.8f", x, y, v);
      fprintf(fplot, "  %12.8f %12.8f", dx_v, dy_v );
      fprintf(fplot, "  %12.8f %12.8f %12.8f", dxx_v, dxy_v, dyy_v);
      fprintf(fplot, "  %12.8f", w);
      fprintf(fplot, "\n"); 
    }

    fprintf(fplot, "\n"); 
  } 
  fclose(fplot);
  fprintf(stderr, "done\n"); 
  free(fname);
}

void interpol_2D
  ( char* method, 
    double x, 
    double y, 
    image_t V, 
    double *v, 
    double *dx_v, 
    double *dy_v, 
    double *dxx_v, 
    double *dxy_v, 
    double *dyy_v, 
    image_t W,
    double *w
  )
{
  if(strcmp(method, "image_eval_STB_new") == 0)
    { /* This method does not give second derivatives. */
      image_eval_STB_new(x, y, V, v, dx_v, dy_v, W, w);
      (*dxx_v) = 0.0; (*dxy_v) = 0.0; (*dyy_v) = 0.0; 
    }
  else if(strcmp(method, "image_eval_STB_old") == 0)
    { /* This method does not give second derivatives. */
      image_eval_STB_old(x, y, V, v, dx_v, dy_v, W, w);
      (*dxx_v) = 0.0; (*dxy_v) = 0.0; (*dyy_v) = 0.0; 
    }
  else
    { fprintf(stderr, "bad method \"%s\"\n", method);
      assert(0);
    }
}


