#include "kltM.h"
#include "lk.h"
#include "lk_error_M.h"
#include "numeric.h"

void klt_find_mark_position_in_image
  ( char *debug_name,
    image_pyramid_t I_pyr,
    image_pyramid_t M_pyr,
    image_pyramid_t W_pyr,  
    int subsampling,  
    double alpha,
    double beta,
    r2_t *p_i,
    r2_t *p_i_dev,
    lk_error_function_t error_func, 
    numeric_minimizer_t minimize,
    double step_factor,
    int niter, 
    double *errorSqr,
    image_pyramid_t I_pyr_deb1,
    image_pyramid_t I_pyr_deb2 ) 
{
  int scale;

  int maxscale = imin(I_pyr->nlevels - 1, M_pyr->nlevels - 1);
  assert(M_pyr->redFactor == I_pyr->redFactor);
  assert(W_pyr->redFactor == I_pyr->redFactor);

  for (scale = maxscale; scale >= 0; scale--) {

    /*Usar uma estrutura auxiliar*/
    r2_t p_i_dev_scaled;
    r2_scale(pow(0.5, scale), p_i_dev, &p_i_dev_scaled);

    char *debug_name_scale = NULL;
    asprintf(&debug_name_scale, "%s_%02d", debug_name, scale);

    /* Map the guessed point {p_i} to the image at level {scale}: */
    if (scale == maxscale) {
      int Inx = I_pyr->imgs[0]->nx;
      int Iny = I_pyr->imgs[0]->ny;
      (*p_i) = image_map_point_to_smaller_image(*p_i, maxscale, I_pyr->redFactor, Inx, Iny);
    }
    else {
      int Inx = I_pyr->imgs[scale]->nx;
      int Iny = I_pyr->imgs[scale]->ny;
      (*p_i) = image_map_point_to_larger_image(*p_i, 1, I_pyr->redFactor, Inx, Iny);
    }
    
    if (I_pyr_deb1 != NULL) {
      /* Put cross at initial guess in {I_pyr_deb1}: */
      image_t I_deb1 = I_pyr_deb1->imgs[scale];
      image_print_cross (I_deb1, 4, *p_i);
    }

    /*ajustar o shape e o pos para a escala*/
    lk_mark_position_adjust
      ( debug_name_scale,
	I_pyr->imgs[scale], 
	M_pyr->imgs[scale], 
	W_pyr->imgs[scale], 
	subsampling,
	alpha,
	beta,
	p_i,
        &p_i_dev_scaled,
        error_func,
        minimize,
        step_factor,
	niter,
	errorSqr
    );

    
    if (I_pyr_deb2 != NULL) {
      /* Put cross at found position in {I_pyr_deb2}: */
      image_t I_deb2 = I_pyr_deb2->imgs[scale];
      image_print_cross (I_deb2, 4, *p_i);
    }
       
    free(debug_name_scale);
  }

}

int klt_choose_mark_scale (int szu, int szv, int subsampling, r2_t *u_i, r2_t *v_i)
{
    double ru = szu/hypot(u_i->c[0], u_i->c[1])/subsampling;
    double rv = szv/hypot(v_i->c[0], v_i->c[1])/subsampling;
    return (int)floor(log(fmax(fmin(ru, rv),1))/M_LN2); 
}
