#ifndef __kltM_H__
#define __kltM_H__

#define _GNU_SOURCE
#include <stdio.h>
#include <malloc.h>
#include <math.h>
#include <image.h>
#include <image_pyramid.h>
#include <jsmath.h>
#include <tf_calib.h>
#include <r2x2.h>
#include <r3.h>
#include <r4.h>
#include <lk.h>

#define DEBUG_WHITE_DOTS_IN_IDEB0 0
/* Put white dots at guess mark positions in image {Ideb0}. */

void klt_find_mark_position_in_image
  ( char *debug_name,
    image_pyramid_t I_pyr,
    image_pyramid_t M_pyr,
    image_pyramid_t W_pyr,  
    int subsampling,  
    double alpha,
    double beta,
    r2_t *p_i,
    r2_t *p_i_dev,
    lk_error_function_t error_func, 
    numeric_minimizer_t minimize,
    double step_factor,
    int niter, 
    double *errorSqr,
    image_pyramid_t I_pyr_deb1,
    image_pyramid_t I_pyr_deb2);
/*
  Locates a mark image {M} in an image {I} by minimizing the quadratic
  discrepancy {Q} defined by {error_func} with a weight mask {W}. It
  uses the multi-scale KLT algorithm, using {niter} iterations of
  Newton's method per level.  The parameters {alpha, beta} are the
  color correction coefficients that must be applied to {M} in order
  to match its occurence in {I}. Requires pyramids for {M,I,W}.

  The location of {M} in {I} is defined by a point {*p_i} which is the
  position of the center of {M} in the domain of {I}, a reduction
  factor {1/subsampling}.  On input, {*p_i} must contain the initial
  estimate for this point. On output, it will contain the optimal
  position found by the procedure; and {*errorSqr} will contain the
  quadratic discrepancy {Q} of the match.
   
  The procedure also draws crosses into the images of {I_pyr_deb1} and
  {I_pyr_deb2} at the initial and final positions {p_i}.

  The string {debug_name} is used to compose file names for debugging. In
  particular, for each scale, the procedure writes the following images ]

     "{debug_name}_{scale}_M.pgm" -- the mark {M}.
     "{debug_name}_{scale}_W.pgm" -- the weight mask {W}.
     "{debug_name}_{scale}_K.pgm" -- the neighborhood of {p_i} in {I} on entry.
     "{debug_name}_{scale}_L.pgm" -- the neighborhood of {p_i} in {I} on exit.

  */

int klt_choose_mark_scale
  ( int szu, 
    int szv, 
    int subsampling, 
    r2_t *u_i, 
    r2_t *v_i );
/* Chooses the scale of a mark based on the {shape} of it, {szu} and {szv} are the size 
 * of the image mark, and {subsampling} is ideal number of subsampling points per pixel*/


#endif
