#include "lk_error_M.h"
#include "lk.h"
#include "numeric.h"


void lk_error_M
  ( image_t I,
    image_t M,
    image_t W,
    int subsampling,
    double alpha,
    double beta, 
    r2_t p_i,
    double *Q,
    double *J, 
    double *H 
 ) 
{
    int k;
    int Mnx = M->nx;
    int Mny = M->ny;
    
    int nz = 2;
    
    int jx, jy;
    double sum_W = 1.0e-100;
    if (Q != NULL) (*Q) = 0.0;
    if (J != NULL) for (k = 0; k < nz; k++) { J[k] = 0.0; }
    if (H != NULL) for (k = 0; k < nz*nz; k++) { H[k] = 0.0; }
    for (jx = 0; jx < M->nx; jx++) {
        for (jy = 0; jy < M->ny; jy++) {
 
            /* Evaluate the mark at the pixel center {gx, gy} (WITHOUT implicit smoothing) */
            double M_val = M->val[jy][jx]; 

            /* Evaluate the image at the corresponding point {px, py} (WITH implicit smoothing) */
            double gx, gy;
            image_pixel_center (jx, jy, &gx, &gy); 
            double px, py;
            image_map_mark_coords_to_image_coords (gx, gy, Mnx, Mny, subsampling, p_i,  &px, &py);
            double I_val, Ix_val, Iy_val, Ixx_val, Ixy_val, Iyy_val;
            image_eval_M (px, py, I, &I_val, &Ix_val, &Iy_val, &Ixx_val, &Ixy_val, &Iyy_val);

            /* Evaluate the weight of this mark pixel */ 
            double W_val = W->val[jy][jx];

            /* Evaluate the color adjusted error term {E_val} */
            double E_val = (alpha * M_val + beta) - I_val;

            double WE = W_val * E_val;

            sum_W += W_val;

            /* Q */
            if (Q != NULL) {
               (*Q) += W_val * E_val*E_val;
            }

            /* J */
            if (J != NULL) { 
               /*deriv(WE/px)*/
               J[0] += -2 * WE * Ix_val;
               /*deriv(WE/py)*/
               J[1] += -2 * WE * Iy_val;
            }

            /* H */
            if (H != NULL) {
               /*deriv(deriv(WE/px),px)*/
               H[0] +=  2 * W_val * pow(Ix_val, 2) - 2 * WE * Ixx_val;
               /*deriv(deriv(WE/px),py)*/
               H[1] +=  2 * W_val * Iy_val * Ix_val - 2 * WE * Ixy_val;
               /*deriv(deriv(WE/py),px)*/
               H[2] +=  2 * W_val * Iy_val * Ix_val - 2 * WE * Ixy_val;
               /*deriv(deriv(WE/py),py)*/
               H[3] +=  2 * W_val * pow(Iy_val, 2) - 2 * WE * Iyy_val;
            }
        }
    }

    if (Q != NULL) (*Q) /= sum_W;
    if (J != NULL) for (k = 0; k < nz; k++) { J[k] /= sum_W; }
    if (H != NULL) for (k = 0; k < nz*nz; k++) { H[k] /= sum_W; }
}

void image_eval_M 
  ( double x, 
    double y, 
    image_t img, 
    double *v, 
    double *dx_v, 
    double *dy_v, 
    double *dxx_v, 
    double *dxy_v, 
    double *dyy_v
  )
{
    /* Adjust (x,y) for the (0.5,0.5) pixel center shift: */
    x -= 0.5;
    y -= 0.5;

    int nx = img->nx;
    int ny = img->ny;

    /* Bicubic interpolation: */
    int ix = floor(x);
    int iy = floor(y);
    double xfrac = x - ix;
    double yfrac = y - iy;
    int t_ok = (v != NULL)||(dx_v != NULL)||(dxx_v != NULL);
    int dy_t_ok = (dy_v != NULL)||(dxy_v != NULL);
    int dyy_t_ok = (dyy_v != NULL);

    int jx;
    double t[4], dy_t[4], dyy_t[4];

    for(jx = ix-1; jx < (ix+3); jx++){
      int ind = jx - ix +1;
      if ((jx >= 0) && (jx < nx)) {
        double p[4];
        p[0] = (((iy-1) >= 0) && ((iy-1) < ny) ? img->val[iy-1][jx]:0);
        p[1] = (((iy+0) >= 0) && ((iy+0) < ny) ? img->val[iy+0][jx]:0);
        p[2] = (((iy+1) >= 0) && ((iy+1) < ny) ? img->val[iy+1][jx]:0);
        p[3] = (((iy+2) >= 0) && ((iy+2) < ny) ? img->val[iy+2][jx]:0);
        interpol_spline_3_B
          ( yfrac, p, 
            (t_ok ? &(t[ind]) : NULL),
            (dy_t_ok ? &(dy_t[ind]) : NULL), 
            (dyy_t_ok ? &(dyy_t[ind]) : NULL)
          );
      } else {
	t[ind] = dy_t[ind] = dyy_t[ind] = 0.0;
      }
    }
    if (t_ok) interpol_spline_3_B (xfrac,t, v, dx_v, dxx_v);
    if (dy_t_ok) interpol_spline_3_B (xfrac,dy_t, dy_v, dxy_v, NULL);
    if (dyy_t_ok) interpol_spline_3_B (xfrac,dyy_t, dyy_v, NULL, NULL);
}
