#ifndef __lk_error_M_H__
#define __lk_error_M_H__

/* R. Minetto's goal function for the Lucas-Kanade algorithm. */

#include <image.h>
#include <r2x2.h>
#include <r3.h>

void lk_error_M
  ( image_t I,
    image_t M,
    image_t W,
    int subsampling,
    double alpha,
    double beta, 
    r2_t p_i,
    double *Q,
    double *J, 
    double *H 
 );
/* Computes the quadratic error {*Q} between the template {M(u,v)}, 
  translated to point {p_i}, and the image {I(x,y)}, weighted by
  {W(u,v)}.

  The parameters {alpha, beta} are the color correction coefficients
  that must be applied to {M} in order to match its occurence in {I}.

  Also computes the gradient {J[0..1]} of {*Q}, and the Hessian
  {H[0..3]} of {*Q}, with respect to {p_i}. The results Q, J and H
  are computed only if the arguments are not NULL*/

void image_eval_M 
  ( double x, 
    double y, 
    image_t img, 
    double *v, 
    double *dx_v, 
    double *dy_v, 
    double *dxx_v, 
    double *dxy_v, 
    double *dyy_v
  );
/* Evaluates the image {img} at the fractional point {x,y}.  Assumes
  that pixel centers are at half-integer coordinates.  The result is
  C1-constinuous but is not an interpolation of the pixel values.
  Returns the result in {*v}.  Also returns the spatial derivatives of
  the value in {*dx_v,*dy_v} and the second derivatives in
  {*dxx_v,*dxy_v,*dyy_v}. */

#endif
