#include "lk_error_STB.h"
#include "lk.h"
#include "numeric.h"
#include <image_STB.h>

void lk_error_STB (image_t I, 
             image_t M, 
             image_t W, 
	     int subsampling,
             double alpha,
             double beta,
             r2_t p_i,
             double *Q,
             double *J,
             double *H) 
{
    int k;
    int Mnx = M->nx;
    int Mny = M->ny;
   
    assert(subsampling == 1); /* Generalize later? */

    int nz = 2;

    int jx, jy;
    if (Q != NULL) (*Q) = 0.0;
    if (J != NULL) for (k = 0; k < nz; k++) { J[k] = 0.0; }
    if (H != NULL) for (k = 0; k < nz*nz; k++) { H[k] = 0.0; }

    double sum_UW = 1.0e-200; 
    for (jx = 0; jx < M->nx; jx++) {
        for (jy = 0; jy < M->ny; jy++) {
            /* Get the mark value {M_val} and its weight {W_val} at the pixel center */
	    double gx, gy; /* Center of pixel {[jy][jx]}. */
            image_pixel_center (jx, jy, &gx, &gy); 
            double M_val, W_val, Mx_val, My_val; 
            image_eval_STB_new (gx, gy, M, &M_val, &Mx_val, &My_val, W, &W_val);
            
            /* Get the image value {I_val} and its weight {U_val} at the corresponding point */
            double px, py; /* Point of {I} corresponding to {gx,gy}. */
            image_map_mark_coords_to_image_coords (gx, gy, Mnx, Mny, subsampling, p_i,  &px, &py);
            double I_val, U_val, Ix_val, Iy_val;
            image_eval_STB_new (px, py, I, &I_val, &Ix_val, &Iy_val, NULL, &U_val);
   
            /* The weight {w} of this pixel in {Q} is the mark weight times the image weight: */
	    double UW = U_val * W_val;
	    sum_UW += UW;
            
            /* Apply color correction to the mark value and gradient: */
            M_val = alpha * M_val + beta;
            Mx_val = alpha * Mx_val;
            My_val = alpha * My_val;

            /* Evaluate the color adjusted error term {E_val} */
            double E_val = M_val - I_val;

            /* Q */
            if (Q != NULL) {
               (*Q) += UW * E_val*E_val;
            }

            /* J */
            if (J != NULL) { 
               J[0] += - 2.0 * UW * E_val * (Mx_val + Ix_val);
               J[1] += - 2.0 * UW * E_val * (My_val + Iy_val);
            }
            
            /* H */
            if (H != NULL) {
               H[0] += UW * pow(Mx_val + Ix_val, 2);
               H[1] += UW * (My_val + Iy_val) * (Mx_val + Ix_val);
               H[2] += UW * (My_val + Iy_val) * (Mx_val + Ix_val);
               H[3] += UW * pow(My_val + Iy_val, 2);
            }
            
	}
    }
    /* Normalize the quadratic error and derivatives to be pixel means: */
    if (Q != NULL) (*Q) /= sum_UW;
    if (J != NULL) for (k = 0; k < nz; k++) { J[k] /= sum_UW; }
    if (H != NULL) for (k = 0; k < nz*nz; k++) { H[k] /= sum_UW; }
}
