#ifndef __lk_error_STB_H__
#define __lk_error_STB_H__

/* Stanley T. Birchfield's goal function for the Lucas-Kanade algorithm. */

#include <image.h>
#include <r2x2.h>
#include <r3.h>

void lk_error_STB
  ( image_t I,
    image_t M,
    image_t W,
    int subsampling,
    double alpha,
    double beta, 
    r2_t p_i,
    double *Q,
    double *J, 
    double *H 
  );
  /* Computes the error {*Q} between the template {M(u,v)}, 
     translated to point {p_i}, and the image {I(x,y)}, weighted by
    {W(u,v)}.

    The parameters {alpha, beta} are the color correction coefficients
    that must be applied to {M} in order to match its occurence in {I}.

    Also computes the gradient {J[0..1]} of {*Q}, and the Hessian
    {H[0..3]} of {*Q}, with respect to {p_i}. The results Q, J and H
    are computed only if the arguments are not NULL. */

#endif
