#include "numeric.h"
#include <assert.h>
#include <rmxn.h>
#include <rn.h>
#include <bool.h>
#include <jsfile.h>

#define DEBUG_ALL 0

void numeric_debug(double x) {
  assert(! isnan(x));
  return;
}

#define WEIGHTED 0
/* If TRUE uses smoothing weights, if FALSE uses plain differences. */

void numeric_diff
  ( double func (int nz, double *z),
    int nz,
    double *z_c,
    double *step, 
    double *J
  )
  {
    int i, k;

    #if (WEIGHTED)
      int nw = 5;
      double w[5] = {-1.0/8.0, -2.0/8.0, 0.0, +2.0/8.0, +1.0/8.0};
    #else 
      int nw = 3;
      double w[3]  ={ -0.5, 00.0, +0.5 };
    #endif


    int hw = nw/2;
    for (i = 0; i < nz; i++) {
        double sum_wQ = 0.0; 
        for (k = 0; k < nw; k++) {
           double coef = k-hw;
           double z_m[nz];
           int t;
           for (t = 0; t < nz; t++) {
               z_m[t] = z_c[t]; 
           } 

           z_m[i] = z_c[i]+coef*step[i];
      
           sum_wQ += w[k]* func (nz, z_m); 
        }
        J[i] = sum_wQ/(step[i]);  
    }
  }



void numeric_diff2
  ( double func (int nz, double *z),
    int nz,
    double *z_c, 
    double *step,
    double *H
  )
  {
    int i, j, k;

    #if (WEIGHTED)
      int nw_1 = 5;
      double w_1[5] = {-1.0/8.0, -2.0/8.0, 0.0, +2.0/8.0, +1.0/8.0};
      int nw_2 = 7;
      double w_2[7] = {
        +0.03819660042851534232,
        +0.18541019576405301903,
        -0.08541018691285015705,
        -0.27639321855943640861,
        -0.08541018691285015705,
        +0.18541019576405301903,
        +0.03819660042851534232
      };
    #else 
      int nw_1 = 3;
      double w_1[3]  ={ -0.5, 00.0, +0.5 };
      int nw_2 = 3;
      double w_2[3]  ={ +1.0, -2.0, +1.0 };
    #endif

    double z_m[nz];

    int hw_1 = nw_1/2;
    int hw_2 = nw_2/2;

    for (i = 0; i < nz; i++) {
        for (j = 0; j < nz; j++) {
             double sum_wQ = 0.0;
             if (i == j) {
                for (k = 0; k < nw_2; k++) {
                    double w = w_2[k];
                    if (w != 0)
                      { double coef = k-hw_2;
                        int t;
                        for (t = 0; t < nz; t++) z_m[t] = z_c[t];
                        z_m[i] = z_c[i] + coef*step[i];
                        sum_wQ += w * func (nz, z_m);
                      } 
                }  
             }
             else {
                int r, s;
                for (r = 0; r < nw_1; r++) {
                    for (s = 0; s < nw_1; s++) {
                       double w = w_1[r]*w_1[s];
                       if (w != 0)
                         { double coef_i = r-hw_1;
                           double coef_j = s-hw_1;
                           int t;
                           for (t = 0; t < nz; t++) z_m[t] = z_c[t]; 
                           z_m[i] = z_c[i] + coef_i*step[i];
                           z_m[j] = z_c[j] + coef_j*step[j];
                           sum_wQ += w * func (nz, z_m);
                         }
                    }  
                }
            }  
            H[nz*i + j] = sum_wQ/(step[i]*step[j]);
        }   
    }
  }

double numeric_norm_vector (double *a, int m) 
  {
    int i; 
    double sum = 0.0;
    for (i = 0; i < m; i++)
        sum += a[i]*a[i];
    return sqrt(sum);
  }

void numeric_plot_all_goal_funcs_and_derivatives
  ( char *name,
    numeric_goal_function_t goal_func, 
    int nz,
    double z[],
    double diff_step[],
    double plot_step[],
    int N
  )
  {
  assert(nz == 2);
  double dza[2] = {+1 ,0}; char *taga = "x"; 
  double dzb[2] = { 0,+1}; char *tagb = "y";
  double dzc[2] = {+1,+1}; char *tage = "x+y";
  double dzd[2] = {+1,-1}; char *tagf = "x-y"; 

  numeric_plot_goal_func_and_derivatives (name, taga, goal_func, nz, z, diff_step, dza, plot_step, N);
  numeric_plot_goal_func_and_derivatives (name, tagb, goal_func, nz, z, diff_step, dzb, plot_step, N);
  numeric_plot_goal_func_and_derivatives (name, tage, goal_func, nz, z, diff_step, dzc, plot_step, N);
  numeric_plot_goal_func_and_derivatives (name, tagf, goal_func, nz, z, diff_step, dzd, plot_step, N);
}

void  numeric_plot_goal_func_and_derivatives   
  ( char *name,
    char *tag,
    numeric_goal_function_t goal_func, 
    int nz,
    double z[],
    double diff_step[],
    double dz[],
    double plot_step[],
    int N)
  {
   int r, k, l;
   double Q;
   double J[nz];
   double J_num[nz];
   double H[nz*nz];
   double H_num[nz*nz];
   double H_mix[nz*nz];
   char *fname = NULL;

   asprintf(&fname, "%s_%s.txt", name, tag);
   FILE *fplot = open_write(fname, TRUE);

   /* Determine stepping direction vector {udir[0..nz-1]}: */
   double udir[nz];   /* Unit vector in stepping direction. */
   double p_mag2 = 0.0;
   for (k = 0; k < nz; k++) {
      udir[k] = dz[k] * plot_step[k];
      p_mag2 += udir[k]*udir[k];
   }
   double p_mag = sqrt(p_mag2); /* Length of plotting step. */
   for (k = 0; k < nz; k++) { udir[k] /= p_mag; }

   for (r = -N; r <= N; r++) {
     double z_m[nz];
     for (k = 0; k < nz; k++) z_m[k] = z[k] + r*p_mag*udir[k]; 

     goal_func (nz, z, &Q, J, H);
     
     numeric_dn_goal_func (goal_func, nz, z, diff_step, J_num);
     numeric_d2n_goal_func (goal_func, nz, z, diff_step, H_num);
     numeric_dn_d_goal_func (goal_func, nz, z, diff_step, H_mix);

     double d_Q = 0.0, dn_Q = 0.0, d2_Q = 0.0, d2n_Q = 0.0, dn_d_Q = 0.0;

     for (k = 0; k < nz; k++) {
         d_Q += J[k]*udir[k];
         dn_Q += J_num[k]*udir[k];
         for (l = 0; l < nz; l++) {
             d2_Q += H[k*nz+l]*udir[k]*udir[l]; 
             d2n_Q += H_num[k*nz+l]*udir[k]*udir[l];
             dn_d_Q += H_mix[k*nz+l]*udir[k]*udir[l];
         }
     }
    
     
     fprintf(fplot, 
       "%12.8f %12.8f  %12.8f %12.8f  %12.8f %12.8f %12.8f ",
       r*p_mag, Q, d_Q, dn_Q, d2_Q, d2n_Q, dn_d_Q);
     for (k = 0; k < nz; k++) { fprintf(fplot, " %12.8f", z_m[k]); }
     fprintf(fplot, "\n"); 

   } 
   fclose(fplot);
   free(fname);
  }

double trim_step(int nz, double z0[], double z1[], double z_min[], double z_max[]) 
  {
    int i;
    /* Find the largest factor {fac <= 1} that leaves the step within the domain: */
    double fac = 1.0; /* Shortening factor to apply. */
    for (i = 0; i < nz; i++) {
      double dzi = z1[i] - z0[i];
      double z_mini = z_min[i]; 
      double z_maxi = z_max[i];
      assert(z0[i] >= z_mini);
      assert(z0[i] <= z_maxi);
      double fi = 1.0;
      if (dzi < 0.0) { fi = (z_mini - z0[i])/dzi; }
      if (dzi > 0.0) { fi = (z_maxi - z0[i])/dzi; }
      /* Reduce {fi} a bit to guard against rounding errors: */
      fi = 0.99999*fi;
      /* Keep the smallest {fi}: */
      fac = fmin(fac, fi);
    }
    if (fac < 1.0) {
      /* Apply {fac} to the step from {z0} to {z1}: */
      for (i = 0; i < nz; i++) { z1[i] = z0[i] + fac*(z1[i] - z0[i]); }
    }
    /* Paranoia check: */
    for (i = 0; i < nz; i++) { assert((z_min[i] <= z1[i]) && (z1[i] <= z_max[i])); }
    return fac;
  }

void numeric_plot_2D_goal_func_and_derivatives
  ( char *outname,
    int nz,
    double z[],
    double plot_step[],
    numeric_goal_function_t goal_func, 
    numeric_minimizer_t minimize, 
    double step_factor,
    double max_disp[],
    int niter,
    double min_step, 
    int N
  )
  {
   assert(nz == 2); 
   int r, s, k, l;
   double Q;
   double J[nz];
   double H[nz*nz];
   char *fname = NULL;

   /* Boundaries of plot area (for Newton convergence test): */
   double z_min[nz], z_max[nz];
   for (k = 0; k < nz; k++) {
     z_min[k] = z[k] - N*plot_step[k] - 0.1;
     z_max[k] = z[k] + N*plot_step[k] + 0.1;
   }

   asprintf(&fname, "%s.txt", outname);
   FILE *fplot = open_write(fname, TRUE);
   for (r = -N; r <= N; r++) {
      if (DEBUG_CONV) { fputc('\n', stderr); }
      for (s = -N; s <= N; s++) {
          double z_m[nz];
          z_m[0] = z[0] + r * plot_step[0]; 
          z_m[1] = z[1] + s * plot_step[1]; 

          goal_func (nz, z_m, &Q, J, H);

          /* Displacements in {X} and {Y} relative to mean position: */
          fprintf(fplot, "%12.8f %12.8f", r*plot_step[0], s*plot_step[1]);
          
          /* Absolute position (packed parameter vector): */
          fprintf(fplot, " ");
          for (k = 0; k < nz; k++) { fprintf(fplot, " %12.8f", z_m[k]); }

          /* Squared error: */
          if (isnan(Q) || (fabs(Q) > 1e10)) { numeric_debug(Q); }
          fprintf(fplot, "  %12.8f", Q);
 
          /* Jacobian: */
          fprintf(fplot, " ");
          for (k = 0; k < nz; k++) {
            if (isnan(J[k]) || (fabs(J[k]) > 1e10)) { numeric_debug(J[k]); }
            fprintf(fplot, " %12.8f", J[k]);
          }

          /* Hessian: */
          fprintf(fplot, " ");
          for (k = 0; k < nz; k++) {
            for (l = 0; l < nz; l++) {
              int r = k*nz + l;
              if (isnan(H[r]) || (fabs(H[r]) > 1e10)) { numeric_debug(H[r]); }
                    fprintf(fplot, " %12.8f", H[r]);
            }
          }

          /* Minimizer's single-step result: */
          double z_u[nz];
          for (k = 0; k < nz; k++) { z_u[k] = z_m[k]; }
          (void) minimize (nz, z_u, goal_func, step_factor, max_disp, 1, min_step);
          
          trim_step(nz, z_m, z_u, z_min, z_max);
          fprintf(fplot, " ");
          for (k = 0; k < nz; k++) fprintf(fplot, " %12.8f", z_u[k]);

          /* Minimizer convergence flag and {niter}-step result: */
          double z_s[nz];
          for (k = 0; k < nz; k++) { z_s[k] = z_m[k]; }
          int converged = minimize(nz, z_s, goal_func, step_factor, max_disp, niter, min_step);
          
          trim_step(nz, z_m, z_s, z_min, z_max);
          fprintf(fplot, "  %d %12.8f %12.8f", converged, z_s[0], z_s[1]);
          
          /* End of data line: */
          fprintf(fplot, "\n"); 
      }

      /* Blank line between scan rows: */
      fprintf(fplot, "\n"); 
   } 
   fclose(fplot);
   free(fname);
  }

void numeric_test_goal_func_derivatives
  ( numeric_goal_function_t goal_func, 
    int nz,
    double z[],
    double diff_step[]
  )
  {
    fprintf(stderr, "--- enter %s ------------------------------\n", __FUNCTION__); 
   int i, j;
   double Q;
   double J[nz];
   double J_num[nz];
   double H[nz*nz];
   double H_num[nz*nz];
   double tol = 1.0e-5;

   goal_func (nz, z, &Q, J, H);

   numeric_dn_goal_func (goal_func, nz, z, diff_step, J_num);

   numeric_d2n_goal_func (goal_func, nz, z, diff_step, H_num);

   for (i = 0; i < nz; i++) {
      for (j = 0; j < nz; j++) {
         double H_diff = H_num[nz*i + j] - H[nz*i + j];
         if (fabs(H_diff) > tol*fabs(Q)) {
	   fprintf(stderr, "H[%d,%d] = %+22.15e, H_num = %+22.15e, diff = %+22.15e\n", i, j, H[nz*i + j], H_num[nz*i + j], H_diff);
         }
      }
      double J_diff = J_num[i] - J[i];
      if (fabs(J_diff) > tol*fabs(Q)) {
	fprintf(stderr, "J[%d] = %+22.15e, J_num = %+22.15e, diff = %+22.15e\n", i, J[i], J_num[i], J_diff);
      }
     
   }

    fprintf(stderr, "--- exit %s -------------------------------\n", __FUNCTION__); 
  }

void numeric_dn_goal_func
  ( numeric_goal_function_t goal_func, 
    int nz, 
    double z[],
    double diff_step[],
    double J[]
  )
  {
    auto double func (int nz, double *z);

    double func (int nz, double *z) {
        double Q;
        goal_func (nz, z, &Q, NULL, NULL); 
        return Q;
    }

    numeric_diff (func, nz, z, diff_step, J);
  }

void numeric_dn_d_goal_func
  ( numeric_goal_function_t goal_func, 
    int nz,
    double z[],
    double diff_step[],
    double H[]
  )
  {
    int i, j;
    double J[nz], H_row[nz];

    for (i = 0; i < nz; i++) {
        auto double func (int nz, double *z);

        double func (int nz, double *z) {
            goal_func (nz, z, NULL, J, NULL);
            return J[i];
        }

        numeric_diff (func, nz, z, diff_step, H_row);
        for (j = 0; j < nz; j++) {
            H[i*nz + j] = H_row[j];
        }
    }
  }

void numeric_d2n_goal_func
  ( numeric_goal_function_t goal_func, 
    int nz,
    double z[],
    double diff_step[], 
    double H[]
   )
  {
    auto double func (int nz, double *z);

    double func (int nz, double *z) {
        double Q; 
        goal_func (nz, z, &Q, NULL, NULL);
        return Q;
    }

    numeric_diff2 (func, nz, z, diff_step, H);
  }

int numeric_Newton_step (int m, double *H, double *J, double *d, double factor)
  {
    double iH[m*m];
    double det = rmxn_inv (m, H, iH);
    if (det == 0.0) {  return 0; }
    rmxn_map_col (m, m, iH, J, d);
    int i;
    for (i = 0; i < m; i++)  d[i] = - factor * d[i];
    return 1;
  }
 
int numeric_iterated_Newton
  ( int nz, 
     double z[], 
     numeric_goal_function_t goal_func,
     double step_factor, 
     double max_disp[], 
     int niter,
     double min_step
   )
   {
     int converged = 1;
     double z_s[nz];
     double z_0[nz];
     double dz_s[nz];
     double Q;
     double J[nz];
     double H[nz*nz];
     /* Set {z_s} to the starting point {z}, save it in {z_0}: */
     int k;
     for (k = 0; k < nz; k++) { z_s[k] = z[k]; z_0[k] = z[k]; }
     int iter = 0; /* Number of minimization steps done. */
     while(1) {
       if (DEBUG_CONV) { fprintf(stderr, "-- Newton's iteration %d\n", iter); }
       if (iter >= niter) { TRACE_CONV('N'); converged = 0; break; }
       /* Compute the gradient and Hessian at {z_s}: */
       goal_func(nz, z_s, &Q, J, H);
       /* Compute the next step {dz_s}: */
       int ok =  numeric_Newton_step(nz, H, J, dz_s, step_factor);
       if (DEBUG_CONV) { fprintf(stderr, " [%+7.3f %+7.3f]", dz_s[0], dz_s[1]); }
       iter++;
       if (! ok) { TRACE_CONV('D'); converged = 0; break; }
       /* Compute {rlen}, {rdisp}: */
       double rlen = 0; /* L-infiniy length of step relative to {max_disp}. */
       double rdisp = 0; /* L-infinity distance from {z} to {z_s}. */
       for (k = 0; k < nz; k++) { 
         z_s[k] += dz_s[k];
         double rlenk = fabs(dz_s[k])/max_disp[k];
         if (rlenk > rlen) { rlen = rlenk; }
         double rdispk = fabs(z_0[k] - z_s[k])/max_disp[k];
         if (rdispk > rdisp) { rdisp = rdispk; }
         if (DEBUG_CONV) { 
           fprintf(stderr, "J[0] = ( %f %f )", J[0], J[1]);
           fprintf(stderr, "  dz_s[%d] = %f", k, dz_s[k]);
           fprintf(stderr, "  rlenk = %f  rdispK = %f  max_disp = %f\n", rlenk, rdispk, max_disp[k]);
	 }
       }
       //rdisp = rlen = 0.0; /*Para executar somente uma iteração*/
       if (rdisp > 1.0) { TRACE_CONV('0' + k); converged = 0; break; }
       /* Here {z_s} is a valid point, even if it will not converge: */
       for (k = 0; k < nz; k++) { 
          z[k] = z_s[k]; 
          if (DEBUG_CONV) { 
            fprintf(stderr, "rlen %f min_step %f novo z[%d] = %f\n", rlen, min_step, k, z[k]);
	  }
       }
       if (rlen < min_step) { TRACE_CONV('!'); converged = 1; break; }
       if (DEBUG_CONV) { 
         fprintf(stderr, "converged = %d   rlen = %f  min_step = %f\n", converged, rlen, min_step); 
       }
     }
     if (DEBUG_CONV) { fprintf(stderr, "final converged = %d\n", converged); }
    return converged;
  }

int numeric_new_safe_Newton
  ( int nz, 
     double z[], 
     numeric_goal_function_t goal_func,
     double step_factor, 
     double max_disp[], 
     int niter,
     double min_step
  )
  {
    int k;
    double dz[nz];
    double z_0[nz]; /*initial guess*/
    double z_s[nz]; /*initial guess*/
    double Q, Q_s; 
    double J[nz];
    double H[nz*nz];

    fprintf(stderr, "--- enter %s ------------------------------\n", __FUNCTION__); 
    /* Save the initial point of this iteration: */
    for (k = 0; k < nz; k++) { z_0[k] = z[k]; }

    int iter = 0;
    while (iter < niter) {
       /* Compute the goal function at {z} */
       goal_func(nz, z, &Q, J, H);

       /* Compute Newton's step: */
       numeric_Newton_step (nz, H, J, dz, step_factor);

       /* Take the step {dz} from {z} to {z_s} (clipped), compute {rlen}: */
       double rlen = 0; /* L-infiniy length of step relative to {max_disp}. */
       for (k = 0; k < nz; k++) { 
         z_s[k] = z[k] + dz[k];
         /* Clip to the specified domain: */
         if (z_s[k] > z_0[k] + max_disp[k]) { z_s[k] = z_0[k] + max_disp[k]; }
         if (z_s[k] < z_0[k] - max_disp[k]) { z_s[k] = z_0[k] - max_disp[k]; }
         /* Compute the relative step length: */
         double rlenk = fabs(z_s[k] - z[k])/max_disp[k];
         if (rlenk > rlen) { rlen = rlenk; }
       }

       /* Try to ensure that the function at {z_s} is not greater than at {z}: */
       while (1) {
             goal_func (nz, z_s, &Q_s, J, H);
             if ((rlen <= min_step) || (Q_s <= Q)) break;
             /* Reflect and shrink {z+s} about {z}: */
             double shrink = 0.5;
             /* fprintf(stderr, "[shrink]\n"); */
             for (k = 0; k < nz; k++) {
                z_s[k] = z[k] - shrink*(z_s[k]-z[k]);
             }
          rlen *= shrink;
       }                    
       if (Q_s > Q) { 
         /* Well, {z} looks like a minimum: */ 
         TRACE_CONV('!'); return 1;
       } else {
         /* Managed to decrease, move to {z_s}: */
         for (k = 0; k < nz; k++) { z[k] = z_s[k]; }
         if (rlen <= min_step) { TRACE_CONV('#'); return 1; }
       } 
       niter++;
    }
    /* Exceeded {niter} iterations: */
    TRACE_CONV('-'); return 0;
  } 


int numeric_STB_safe_Newton
  ( int nz, 
     double z[], 
     numeric_goal_function_t goal_func,
     double step_factor, 
     double max_disp[], 
     int niter,
     double min_step
   )
{
   double Q; /* function value*/
   double J[nz]; /*Jacobian*/
   double H[nz*nz]; /*Hessian*/
   double Dmin = 0.01; /* minimum step on each iteration */
   double Dmax = 1.0; /* search half-radius around the original parameters*/

   goal_func(nz, z, &Q, J, H);
   
   int iter = 0;
   while (iter < niter) {
       double D = numeric_STB_safe_Newton_step(goal_func, z, nz, &Q, J, H, Dmax);
       if ((iter > 0) && (D < Dmin)) { TRACE_CONV('!'); return 1; }
       Dmax /= 2.0;
       iter++;
   }
   TRACE_CONV('-'); return 0;
}

double numeric_STB_safe_Newton_step
  ( numeric_goal_function_t goal_func, 
    double *z, 
    int nz, 
    double *Q, 
    double *J, 
    double *H, 
    double Dmax
  )
{
       int k;
       double dz[nz];
       double z0[nz]; /*initial guess*/
       double Q0; /*initial guess*/
       double J0[nz]; /*initial guess*/
       double H0[nz*nz]; /*initial guess*/

       /* saving the initial parameters */
       Q0 = *Q;
       for (k = 0; k < nz; k++) {
           z0[k] = z[k];
           J0[k] = J[k];
       }
       for (k = 0; k < nz*nz; k++) H0[k] = H[k];

       numeric_Newton_step (nz, H, J, dz, 1.0);

       /* Compute distance {D} from previous parameters */
       double D2 = 0.0; /*distance squared*/
       for (k = 0; k < nz; k++) {
          z[k] += dz[k];
          double Dk =  z[k] - z0[k];
          D2 += Dk*Dk;
       }
       double D = sqrt(D2);

       /*prevent excessive displacement*/
       if (D > Dmax) {
            /* fprintf(stderr, "[trim]\n"); */
            for (k = 0; k < nz; k++) {
               z[k] = z0[k] + (Dmax/D)*(z[k]-z0[k]); 
           }
           D = Dmax;
       }

       int ntries = 3; 

       while (1) {
          goal_func (nz, z, Q, J, H);

          ntries--;
          if ((ntries == 0) || (*Q <= Q0)) break;
          /* reflect and shrink z about z0 */
          double shrink = 0.5;
          /* fprintf(stderr, "[shrink]\n"); */
          for (k = 0; k < nz; k++) {
             z[k] = z0[k] - shrink*(z[k]-z0[k]);
          }
          D = shrink*D;
      }          
      if (*Q <= Q0) return D;
       

      /* Restore the original values */
      *Q = Q0;
       for (k = 0; k < nz; k++) {
           z[k] = z0[k];
           J[k] = J0[k];
       }
       for (k = 0; k < nz*nz; k++) H[k] = H0[k];
      return 0.0;
}

