#ifndef __numeric_min_nwgrad_H__
#define __numeric_min_nwgrad_H__

#define _GNU_SOURCE
#include <stdio.h>
#include <rn.h>
#include <malloc.h>
#include <math.h>
#include <jsmath.h>
#include "numeric.h"

int numeric_grad_Newton
  ( int nz, 
     double z[], 
     numeric_goal_function_t goal_func,
     double step_factor, 
     double max_disp[], 
     int niter,
     double min_step
  );
  /* A combination of iterated Newton and gradient descent.  At each
    iteration, applies {numeric_grad_Newton_step} to the current guess
    {z[0..nz-1]}. */

int numeric_grad_Newton_step
  ( int nz, 
    double z[], 
    numeric_goal_function_t goal_func,
    double *Q, 
    double *J, 
    double *H,
    double step_factor, 
    double z_min[], 
    double z_max[],
    double grad_step, 
    double min_step
  );
 /* The main iteration of {numeric_grad_Newton}.  It tries to 
   improve the point {z[0..nz-1]} (which is assumed to have the
   given {Q,J,H} values) by one of several strategies. 

   First, it applies {numeric_Newton_step} (reduced by {step_factor})
   to the current point {z[0..nz-1]}. 

   If that fails, it tries gradient descent along the original
   gradient {J} with various step lengths, starting with {grad_step}
   and decreasing until the step length becomes less than {min_step}.

   In each case, the step length is trimmed so that {z} remains inside
   the region defined by {z_min[0..nz-1]} and {z_max[0..nz-1]}. The
   starting point {z} must be inside that region.

   If any of these attempted steps succeeds in decreasing the value
   {Q}, the procedure moves {z} to the end of that step, updates
   {Q,J,H}, and returns 1 ("improved").

   If all these attempts fail to improve {Q}, the procedure returns 0
   ("failed to improve") without changing {Q,J,H}.  */

#endif
