
void tf_tsai (calibration_data_t cdat, camera_parameters_t cpar, double wgts[], FILE *ferr)
{
    rm_t U;
    r3_t R; /*Rotation angles*/
    calibration_aux_data_t caux;

    if (wgts == NULL) {
        fprintf(stderr, "######################################################################################################################################################################################################################################################################## \nSem pesos\n");
    }

    caux = tf_create_auxiliary_calibration_aux_data_structure (cdat->ntargets);

    tf_compute_xd_yd_and_r_squared (cdat, cpar, caux);

    U = tf_compute_U (cdat, caux, wgts, ferr);

    tf_compute_tx_and_ty (cdat, cpar, caux, U);

    tf_compute_sx (cpar, U);
    
    tf_compute_xd_yd_and_r_squared (cdat, cpar, caux);

    tf_compute_better_R (cpar, U, &R);

    tf_compute_approximate_f_and_Tz (cpar, cdat, caux, wgts);

    if (ferr != NULL) {
         fprintf(ferr, "Sx : %f, Tx : %f, Ty : %f\n", cpar->sx, cpar->S.c[1][0], cpar->S.c[2][0]);
         fprintf(ferr, "r1 : %f, r2 : %f, r3 : %f\n", cpar->S.c[1][1], cpar->S.c[1][2], cpar->S.c[1][3]);
         fprintf(ferr, "r4 : %f, r5 : %f, r6 : %f\n", cpar->S.c[2][1], cpar->S.c[2][2], cpar->S.c[2][3]);
         fprintf(ferr, "r7 : %f, r8 : %f, r9 : %f\n", cpar->S.c[3][1], cpar->S.c[3][2], cpar->S.c[3][3]);
         fprintf(ferr, "u1 : %f, u2 : %f, u3 : %f, u4 : %f\n", U->c[0], U->c[1], U->c[2], U->c[3]);
         fprintf(ferr, "u5 : %f, u6 : %f, u7 : %f\n", U->c[4], U->c[5], U->c[6]);
    }

    if (cpar->f < 0) {
        /*try the other solution for the orthonormal matrix*/
        cpar->S.c[1][3] = -cpar->S.c[1][3]; /*r3*/
        cpar->S.c[2][3] = -cpar->S.c[2][3]; /*r6*/
        cpar->S.c[3][1] = -cpar->S.c[3][1]; /*r7*/
        cpar->S.c[3][2] = -cpar->S.c[3][2]; /*r8*/

        tf_solve_RPY_transform (cpar, &R);

        tf_compute_approximate_f_and_Tz (cpar, cdat, caux, wgts);

        if (cpar->f < 0) {
            fprintf (stderr, "error - possible handedness problem with data\n");
            //exit (-1);
        }
    } 

    /*try the exact solution by lmdif*/
    tf_compute_exact_f_and_Tz (cpar, cdat, caux, wgts);

    /* do a full optimization minus the image center */
    ncc_nic_optimization (&R, cdat, cpar);

    /* do a full optimization including the image center */
    ncc_full_optimization (&R, cdat, cpar);

    /*write the parameters calculated by Tsai*/
    //tf_print_error_stats (cdat, cpar, ferr);

    /*disallocating structures*/
    tf_free_rm_structure (U);
    tf_free_auxiliary_calibration_aux_data_structure (caux);
}
