
void tf_compute_xd_yd_and_r_squared (calibration_data_t cdat, camera_parameters_t cpar, calibration_aux_data_t caux)
{   
    int i;

    for (i = 0; i < cdat->ntargets; i++) {
        caux->distorted_coords[i] = tf_image_coords_to_sensor_coords (cpar, cdat->image_coords[i]);
 
        /*r_squared*/
        caux->r_squared[i] = SQR (caux->distorted_coords[i].c[0]) + SQR (caux->distorted_coords[i].c[1]);
    }
}

rm_t tf_compute_U (calibration_data_t cdat, calibration_aux_data_t caux, double wgts[], FILE *ferr)
{
    int i;
    mat_rm_t M = tf_alloc_mat_rm (cdat->ntargets, 7);
    rm_t b = tf_alloc_rm (cdat->ntargets);
    rm_t U = tf_alloc_rm (7);

    for (i = 0; i < cdat->ntargets; i++) {
        M->c[i*7]   =  caux->distorted_coords[i].c[1] * cdat->world_coords[i].c[0];
        M->c[i*7+1] =  caux->distorted_coords[i].c[1] * cdat->world_coords[i].c[1];
        M->c[i*7+2] =  caux->distorted_coords[i].c[1] * cdat->world_coords[i].c[2];
        M->c[i*7+3] =  caux->distorted_coords[i].c[1];
        M->c[i*7+4] = -caux->distorted_coords[i].c[0] * cdat->world_coords[i].c[0];
        M->c[i*7+5] = -caux->distorted_coords[i].c[0] * cdat->world_coords[i].c[1];
        M->c[i*7+6] = -caux->distorted_coords[i].c[0] * cdat->world_coords[i].c[2];
        b->c[i]     =  caux->distorted_coords[i].c[0];
	fprintf(ferr, "Dist : %d, %f, %f\n", i, caux->distorted_coords[i].c[0], caux->distorted_coords[i].c[1]);
	fprintf(ferr, "World: %f, %f, %f\n", cdat->world_coords[i].c[0], cdat->world_coords[i].c[1], cdat->world_coords[i].c[2]);
    }

    tf_solve_system_mxn (M, U, b, wgts);

    tf_free_mat_rm_structure (M);
    tf_free_rm_structure (b);

    return U;
}

void tf_compute_tx_and_ty (calibration_data_t cdat, camera_parameters_t cpar, calibration_aux_data_t caux, rm_t U)
{
    int i, far_point;
    double Tx, Ty, Ty_squared, x, y;
    double distance, far_distance;
    
    /* first find the square of the magnitude of Ty */
    Ty_squared = 1 / (SQR (U->c[4]) + SQR (U->c[5]) + SQR (U->c[6]));

    /* find a point that is far from the image center */
    far_distance = 0;
    far_point = 0;
    for (i = 0; i < cdat->ntargets; i++) {
        if ((distance = caux->r_squared[i]) > far_distance) {
            far_point = i;
            far_distance = distance;
        }
    }  
    
    /* now find the sign for Ty, start by assuming Ty > 0 */
    Ty = sqrt (Ty_squared);
    Tx = U->c[3] * Ty;

    x = (U->c[0] * Ty) * cdat->world_coords[far_point].c[0] + 
        (U->c[1] * Ty) * cdat->world_coords[far_point].c[1] +
        (U->c[2] * Ty) * cdat->world_coords[far_point].c[2] + Tx;

    y = (U->c[4] * Ty) * cdat->world_coords[far_point].c[0] + 
        (U->c[5] * Ty) * cdat->world_coords[far_point].c[1] +
        (U->c[6] * Ty) * cdat->world_coords[far_point].c[2] + Ty;

    /* flip Ty if we guessed wrong */
    if ( (SIGNBIT (x) != SIGNBIT (caux->distorted_coords[far_point].c[0])) || 
         (SIGNBIT (y) != SIGNBIT (caux->distorted_coords[far_point].c[1])) )
         Ty = -Ty;

    /* update the calibration constants */
    cpar->S.c[1][0] = U->c[3] * Ty;    /*Tx*/
    cpar->S.c[2][0] = Ty;              /*Ty*/
}

void tf_compute_sx (camera_parameters_t cpar, rm_t U)
{
    cpar->sx = sqrt (SQR (U->c[0]) + SQR (U->c[1]) + SQR (U->c[2])) * fabs (cpar->S.c[2][0]);
}

void tf_compute_better_R (camera_parameters_t cpar, rm_t U, r3_t *R)
{
    double r1, r2, r3, r4, r5, r6, r7;
    double sa, ca, sb, cb, sg, cg;

    r1 = U->c[0] * cpar->S.c[2][0] / cpar->sx;
    r2 = U->c[1] * cpar->S.c[2][0] / cpar->sx;
    r3 = U->c[2] * cpar->S.c[2][0] / cpar->sx;
    r4 = U->c[4] * cpar->S.c[2][0];
    r5 = U->c[5] * cpar->S.c[2][0];
    r6 = U->c[6] * cpar->S.c[2][0];

    /* use the outer product of the first two rows to get the last row */
    r7 = r2 * r6 - r3 * r5;

    /* now find the RPY angles corresponding to the estimated rotation matrix */
    R->c[2] = atan2 (r4, r1);

    SINCOS (R->c[2], sg, cg);

    R->c[1] = atan2 (-r7, r1 * cg + r4 * sg);
    
    R->c[0] = atan2 (r3 * sg - r6 * cg, r5 * cg - r2 * sg);

    SINCOS (R->c[0], sa, ca);
    
    SINCOS (R->c[1], sb, cb);

    /* now generate a more orthonormal rotation matrix from the RPY angles */
    cpar->S.c[1][1] =  cb * cg;                  /*r1*/      
    cpar->S.c[1][2] =  cg * sa * sb - ca * sg;   /*r2*/
    cpar->S.c[1][3] =  sa * sg + ca * cg * sb;   /*r3*/
    cpar->S.c[2][1] =  cb * sg;                  /*r4*/ 
    cpar->S.c[2][2] =  sa * sb * sg + ca * cg;   /*r5*/
    cpar->S.c[2][3] =  ca * sb * sg - cg * sa;   /*r6*/ 
    cpar->S.c[3][1] = -sb;                       /*r7*/
    cpar->S.c[3][2] =  cb * sa;                  /*r8*/
    cpar->S.c[3][3] =  ca * cb;                  /*r9*/
}

void tf_solve_RPY_transform (camera_parameters_t cpar, r3_t *R)
{
    double    sg, cg;

    R->c[2] = atan2 (cpar->S.c[2][1], cpar->S.c[1][1]);

    SINCOS (R->c[2], sg, cg);

    R->c[1] = atan2 (-cpar->S.c[3][1], cpar->S.c[1][1] * cg + cpar->S.c[2][1] * sg);

    R->c[0] = atan2 (cpar->S.c[1][3] * sg - cpar->S.c[2][3] * cg, cpar->S.c[2][2] * cg - cpar->S.c[1][2] * sg);
}

void tf_compute_approximate_f_and_Tz (camera_parameters_t cpar, calibration_data_t cdat, calibration_aux_data_t caux, double wgts[])
{
    int i;
    mat_rm_t M = tf_alloc_mat_rm (cdat->ntargets, 2);
    rm_t b = tf_alloc_rm (cdat->ntargets);
    rm_t U = tf_alloc_rm (2);

    for (i = 0; i < cdat->ntargets; i++) {

        M->c[i*2] =  cpar->S.c[2][1] * cdat->world_coords[i].c[0] + 
                     cpar->S.c[2][2] * cdat->world_coords[i].c[1] + 
                     cpar->S.c[2][3] * cdat->world_coords[i].c[2] + cpar->S.c[2][0];

        M->c[i*2+1] = -caux->distorted_coords[i].c[1];

        b->c[i]  = ( cpar->S.c[3][1] * cdat->world_coords[i].c[0] + 
                     cpar->S.c[3][2] * cdat->world_coords[i].c[1] + 
                     cpar->S.c[3][3] * cdat->world_coords[i].c[2] ) * caux->distorted_coords[i].c[1];
    }

    tf_solve_system_mxn (M, U, b, wgts);

    /* update the calibration constants */
    cpar->f = U->c[0];
    cpar->S.c[3][0] = U->c[1]; /*Tz*/
    cpar->kappa = 0.0; /* this is the assumption that our calculation was made under */

    tf_free_mat_rm_structure (M);
    tf_free_rm_structure (b); 
    tf_free_rm_structure (U); 
}

void tf_compute_exact_f_and_Tz (camera_parameters_t cpar, calibration_data_t cdat, calibration_aux_data_t caux, double wgts[])
{
    #define NPARAMS 3

    int       i;

    /* Parameters needed by MINPACK's lmdif() */

    integer     m = cdat->ntargets;
    integer     n = NPARAMS;
    doublereal  x[NPARAMS];
    doublereal *fvec;
    doublereal  ftol = REL_SENSOR_TOLERANCE_ftol;
    doublereal  xtol = REL_PARAM_TOLERANCE_xtol;
    doublereal  gtol = ORTHO_TOLERANCE_gtol;
    integer     maxfev = MAXFEV;
    doublereal  epsfcn = EPSFCN;
    doublereal  diag[NPARAMS];
    integer     mode = MODE;
    doublereal  factor = FACTOR;
    integer     nprint = 0;
    integer     info;
    integer     nfev;
    doublereal *fjac;
    integer     ldfjac = m;
    integer     ipvt[NPARAMS];
    doublereal  qtf[NPARAMS];
    doublereal  wa1[NPARAMS];
    doublereal  wa2[NPARAMS];
    doublereal  wa3[NPARAMS];
    doublereal *wa4;

    /* allocate some workspace */
    if (( fvec = (doublereal *) calloc ((unsigned int) m, (unsigned int) sizeof(doublereal))) == NULL ) {
       fprintf(stderr,"calloc: Cannot allocate workspace fvec\n");
       exit(-1);
    }

    if (( fjac = (doublereal *) calloc ((unsigned int) m*n, (unsigned int) sizeof(doublereal))) == NULL ) {
       fprintf(stderr,"calloc: Cannot allocate workspace fjac\n");
       exit(-1);
    }

    if (( wa4 = (doublereal *) calloc ((unsigned int) m, (unsigned int) sizeof(doublereal))) == NULL ) {
       fprintf(stderr,"calloc: Cannot allocate workspace wa4\n");
       exit(-1);
    }

    /* use the current calibration constants as an initial guess */
    x[0] = cpar->f;
    x[1] = cpar->S.c[3][0]; /*Tz*/
    x[2] = cpar->kappa;

    /* FUNCTIONS TO PERFORM DEBUG
    tf_print_calibration_parameters (cpar, stdout);
    tf_print_weights (cdat->ntargets, wgts, stdout); 
    tf_print_calibration_aux_data (caux, stdout);
    tf_print_calibration_data (cdat, stdout);*/

    /* define optional scale factors for the parameters */
    if ( mode == 2 ) {
        for (i = 0; i < NPARAMS; i++)
            diag[i] = 1.0; /* some user-defined values */
    }

    /* perform the optimization */
    doublereal uwerr[m];

    auto void tf_compute_error (integer *mp, integer *np, doublereal *x, doublereal *fvec, integer *iflag);

    void tf_compute_error (integer *mp, integer *np, doublereal *x, doublereal *fvec, integer *iflag) {
         assert(*mp == m); assert(*np == n);
         tf_compute_exact_f_and_Tz_error(x, fvec, uwerr, wgts, cdat, cpar, caux);
    }

    lmdif  (tf_compute_error, &m, &n, x, fvec, &ftol, &xtol, &gtol, &maxfev, &epsfcn, diag, &mode, 
            &factor, &nprint, &info, &nfev, fjac, &ldfjac, ipvt, qtf, wa1, wa2, wa3, wa4);

    tf_compute_exact_f_and_Tz_error (x, fvec, uwerr, wgts, cdat, cpar, caux);

    /* update the calibration constants */
    cpar->f = x[0];
    cpar->S.c[3][0] = x[1];
    cpar->kappa = x[2];

    /* release allocated workspace */
    free (fvec);
    free (fjac);
    free (wa4);
    #undef NPARAMS
}

void  tf_compute_exact_f_and_Tz_error (doublereal *params, doublereal *err,  doublereal *uwerr, double wgts[], 
                                       calibration_data_t cdat, camera_parameters_t cpar, calibration_aux_data_t caux)
{
    int       i;

    double    xc, yc, zc, Xu_1, Yu_1, Xu_2, Yu_2, distortion_factor, f, Tz, kappa;

    f = params[0];
    Tz = params[1];
    kappa = params[2];

    for (i = 0; i < cdat->ntargets; i++) {

        /* convert from world coordinates to camera coordinates */
        xc = cpar->S.c[1][1] * cdat->world_coords[i].c[0] + 
             cpar->S.c[1][2] * cdat->world_coords[i].c[1] + 
             cpar->S.c[1][3] * cdat->world_coords[i].c[2] + cpar->S.c[1][0];

        yc = cpar->S.c[2][1] * cdat->world_coords[i].c[0] + 
             cpar->S.c[2][2] * cdat->world_coords[i].c[1] + 
             cpar->S.c[2][3] * cdat->world_coords[i].c[2] + cpar->S.c[2][0];

        zc = cpar->S.c[3][1] * cdat->world_coords[i].c[0] + 
             cpar->S.c[3][2] * cdat->world_coords[i].c[1] + 
             cpar->S.c[3][3] * cdat->world_coords[i].c[2] + Tz;

        /* convert from camera coordinates to undistorted sensor coordinates */
        Xu_1 = f * xc / zc;
        Yu_1 = f * yc / zc;

        /* convert from distorted sensor coordinates to undistorted sensor coordinates */
        distortion_factor = 1 + kappa * caux->r_squared[i];
        Xu_2 = caux->distorted_coords[i].c[0] * distortion_factor;
        Yu_2 = caux->distorted_coords[i].c[1] * distortion_factor;

        /* record the error in the undistorted sensor coordinates */
        uwerr[i] = hypot (Xu_1 - Xu_2, Yu_1 - Yu_2);
        double wti = (wgts == NULL ? 1.0 : wgts[i]);
        err[i] = sqrt(wti) * uwerr[i];
    }
}



/************************************************************************/

void  ncc_nic_optimization_error (  integer *m_ptr, 
                                    integer *n_ptr, 
				    doublereal *params, 
				    doublereal *err,
				    calibration_data_t cdat, 
				    camera_parameters_t cpar )
{
    int       i;

    double    xc, yc, zc;
    double    Xd_, Yd_, Xu_1, Yu_1, Xu_2, Yu_2;
    double    distortion_factor;
    double    Rx, Ry, Rz, Tx, Ty, Tz;
    double    kappa1, sx, f;
    double    r1, r2, r3, r4, r5, r6, r7, r8, r9;
    double    sa, sb, sg, ca, cb, cg;

    Rx = params[0];
    Ry = params[1];
    Rz = params[2];
    Tx = params[3];
    Ty = params[4];
    Tz = params[5];
    kappa1 = params[6];
    f = params[7];
    sx = params[8];

    SINCOS (Rx, sa, ca);
    SINCOS (Ry, sb, cb);
    SINCOS (Rz, sg, cg);
    r1 = cb * cg;
    r2 = cg * sa * sb - ca * sg;
    r3 = sa * sg + ca * cg * sb;
    r4 = cb * sg;
    r5 = sa * sb * sg + ca * cg;
    r6 = ca * sb * sg - cg * sa;
    r7 = -sb;
    r8 = cb * sa;
    r9 = ca * cb;

    for (i = 0; i < cdat->ntargets; i++) {
        // convert from world coordinates to camera coordinates 
    
        xc = r1 * cdat->world_coords[i].c[0] + r2 * cdat->world_coords[i].c[1] + r3 * cdat->world_coords[i].c[2] + Tx;
        yc = r4 * cdat->world_coords[i].c[0] + r5 * cdat->world_coords[i].c[1] + r6 * cdat->world_coords[i].c[2] + Ty;
        zc = r7 * cdat->world_coords[i].c[0] + r8 * cdat->world_coords[i].c[1] + r9 * cdat->world_coords[i].c[2] + Tz;

        // convert from camera coordinates to undistorted sensor plane coordinates
        Xu_1 = f * xc / zc;
        Yu_1 = f * yc / zc;

        // convert from 2D image coordinates to distorted sensor coordinates 
        Xd_ = cpar->dpx * (cdat->image_coords[i].c[0] - cpar->Cx) / sx;
        Yd_ = cpar->dpy * (cdat->image_coords[i].c[1] - cpar->Cy);

        // convert from distorted sensor coordinates to undistorted sensor plane coordinates 
        distortion_factor = 1 + kappa1 * (SQR (Xd_) + SQR (Yd_));
        Xu_2 = Xd_ * distortion_factor;
        Yu_2 = Yd_ * distortion_factor;

        // record the error in the undistorted sensor coordinates 
        err[i] = hypot (Xu_1 - Xu_2, Yu_1 - Yu_2);
    }

}

void      ncc_nic_optimization (  r3_t *R,
                                  calibration_data_t cdat,
                                  camera_parameters_t cpar )
{
#define NPARAMS 9

    int       i;

    /* Parameters needed by MINPACK's lmdif() */

    integer     m = cdat->ntargets;
    integer     n = NPARAMS;
    doublereal  x[NPARAMS];
    doublereal *fvec;
    doublereal  ftol = REL_SENSOR_TOLERANCE_ftol;
    doublereal  xtol = REL_PARAM_TOLERANCE_xtol;
    doublereal  gtol = ORTHO_TOLERANCE_gtol;
    integer     maxfev = MAXFEV;
    doublereal  epsfcn = EPSFCN;
    doublereal  diag[NPARAMS];
    integer     mode = MODE;
    doublereal  factor = FACTOR;
    integer     nprint = 0;
    integer     info;
    integer     nfev;
    doublereal *fjac;
    integer     ldfjac = m;
    integer     ipvt[NPARAMS];
    doublereal  qtf[NPARAMS];
    doublereal  wa1[NPARAMS];
    doublereal  wa2[NPARAMS];
    doublereal  wa3[NPARAMS];
    doublereal *wa4;

     /* allocate some workspace */
    if (( fvec = (doublereal *) calloc ((unsigned int) m, (unsigned int) sizeof(doublereal))) == NULL ) {
       fprintf(stderr,"calloc: Cannot allocate workspace fvec\n");
       exit(-1);
    }

    if (( fjac = (doublereal *) calloc ((unsigned int) m*n, (unsigned int) sizeof(doublereal))) == NULL ) {
       fprintf(stderr,"calloc: Cannot allocate workspace fjac\n");
       exit(-1);
    }

    if (( wa4 = (doublereal *) calloc ((unsigned int) m, (unsigned int) sizeof(doublereal))) == NULL ) {
       fprintf(stderr,"calloc: Cannot allocate workspace wa4\n");
       exit(-1);
    }

    /* use the current calibration and camera constants as a starting point */
    x[0] = R->c[0];
    x[1] = R->c[1];
    x[2] = R->c[2];
    x[3] = cpar->S.c[1][0];
    x[4] = cpar->S.c[2][0];
    x[5] = cpar->S.c[3][0];
    x[6] = cpar->kappa;
    x[7] = cpar->f;
    x[8] = cpar->sx;

        /* define optional scale factors for the parameters */
    if ( mode == 2 ) {
        for (i = 0; i < NPARAMS; i++)
            diag[i] = 1.0;             /* some user-defined values */
    }

    auto void tf_ncc_nic_optimization_error (integer *m_ptr, integer *n_ptr, doublereal *params, doublereal *err, integer *iflag);
    
    void tf_ncc_nic_optimization_error (integer *m_ptr, integer *n_ptr, doublereal *params, doublereal *err, integer *iflag) {
         assert(*m_ptr == m); assert(*n_ptr == n);
         ncc_nic_optimization_error (m_ptr, n_ptr, x, fvec, cdat, cpar); 
    }



    /* perform the optimization */
    lmdif  (tf_ncc_nic_optimization_error, &m, &n, x, fvec, &ftol, &xtol, &gtol, &maxfev, &epsfcn,
            diag, &mode, &factor, &nprint, &info, &nfev, fjac, &ldfjac, ipvt, qtf, wa1, wa2, wa3, wa4);

    /* update the calibration and camera constants */
    R->c[0] = x[0];
    R->c[1] = x[1];
    R->c[2] = x[2];
    tf_solve_RPY_transform (cpar, R);

    cpar->S.c[1][0] = x[3];
    cpar->S.c[2][0] = x[4];
    cpar->S.c[3][0] = x[5];
    cpar->kappa = x[6];
    cpar->f = x[7];
    cpar->sx = x[8];

    /* release allocated workspace */
    free (fvec);
    free (fjac);
    free (wa4);

#ifdef DEBUG
    /* print the number of function calls during iteration */
    fprintf(stderr,"info: %d nfev: %d\n\n",info,nfev);
#endif

#undef NPARAMS
}

void      ncc_full_optimization_error ( integer *m_ptr, 
                                        integer *n_ptr, 
                                        doublereal *params, 
                                        doublereal *err,
                                        calibration_data_t cdat,
                                        camera_parameters_t cpar  )
{
    int i;

    double xc, yc, zc;
    double Xd_, Yd_, Xu_1, Yu_1, Xu_2, Yu_2;
    double distortion_factor;
    double Rx, Ry, Rz;
    double Tx, Ty, Tz;
    double kappa1, sx, f;
    double Cx, Cy;
    double r1, r2, r3, r4, r5, r6, r7, r8, r9;
    double sa, sb, sg, ca, cb, cg;

    Rx = params[0];
    Ry = params[1];
    Rz = params[2];
    Tx = params[3];
    Ty = params[4];
    Tz = params[5];
    kappa1 = params[6];
    f = params[7];
    sx = params[8];
    Cx = params[9];
    Cy = params[10];

    SINCOS (Rx, sa, ca);
    SINCOS (Ry, sb, cb);
    SINCOS (Rz, sg, cg);
    r1 = cb * cg;
    r2 = cg * sa * sb - ca * sg;
    r3 = sa * sg + ca * cg * sb;
    r4 = cb * sg;
    r5 = sa * sb * sg + ca * cg;
    r6 = ca * sb * sg - cg * sa;
    r7 = -sb;
    r8 = cb * sa;
    r9 = ca * cb;

    for (i = 0; i < cdat->ntargets; i++) {
        /* convert from world coordinates to camera coordinates */
        xc = r1 * cdat->world_coords[i].c[0] + r2 * cdat->world_coords[i].c[1] + r3 * cdat->world_coords[i].c[2] + Tx;
        yc = r4 * cdat->world_coords[i].c[0] + r5 * cdat->world_coords[i].c[1] + r6 * cdat->world_coords[i].c[2] + Ty;
        zc = r7 * cdat->world_coords[i].c[0] + r8 * cdat->world_coords[i].c[1] + r9 * cdat->world_coords[i].c[2] + Tz;

        /* convert from camera coordinates to undistorted sensor plane coordinates */
        Xu_1 = f * xc / zc;
        Yu_1 = f * yc / zc;

        /* convert from 2D image coordinates to distorted sensor coordinates */
        Xd_ = cpar->dpx * (cdat->image_coords[i].c[0] - cpar->Cx) / sx;
        Yd_ = cpar->dpy * (cdat->image_coords[i].c[1] - cpar->Cy);

        /* convert from distorted sensor coordinates to undistorted sensor plane coordinates */
        distortion_factor = 1 + kappa1 * (SQR (Xd_) + SQR (Yd_));
        Xu_2 = Xd_ * distortion_factor;
        Yu_2 = Yd_ * distortion_factor;

        /* record the error in the undistorted sensor coordinates */
        err[i] = hypot (Xu_1 - Xu_2, Yu_1 - Yu_2);
    }
}
 
void      ncc_full_optimization (  r3_t *R,
                                   calibration_data_t cdat,
                                   camera_parameters_t cpar ) 
{
#define NPARAMS 11

    int     i;

    /* Parameters needed by MINPACK's lmdif() */

    integer     m = cdat->ntargets;
    integer     n = NPARAMS;
    doublereal  x[NPARAMS];
    doublereal *fvec;
    doublereal  ftol = REL_SENSOR_TOLERANCE_ftol;
    doublereal  xtol = REL_PARAM_TOLERANCE_xtol;
    doublereal  gtol = ORTHO_TOLERANCE_gtol;
    integer     maxfev = MAXFEV;
    doublereal  epsfcn = EPSFCN;
    doublereal  diag[NPARAMS];
    integer     mode = MODE;
    doublereal  factor = FACTOR;
    integer     nprint = 0;
    integer     info;
    integer     nfev;
    doublereal *fjac;
    integer     ldfjac = m;
    integer     ipvt[NPARAMS];
    doublereal  qtf[NPARAMS];
    doublereal  wa1[NPARAMS];
    doublereal  wa2[NPARAMS];
    doublereal  wa3[NPARAMS];
    doublereal *wa4;
    /* allocate some workspace */
    if (( fvec = (doublereal *) calloc ((unsigned int) m, (unsigned int) sizeof(doublereal))) == NULL ) {
       fprintf(stderr,"calloc: Cannot allocate workspace fvec\n");
       exit(-1);
    }

    if (( fjac = (doublereal *) calloc ((unsigned int) m*n, (unsigned int) sizeof(doublereal))) == NULL ) {
       fprintf(stderr,"calloc: Cannot allocate workspace fjac\n");
       exit(-1);
    }

    if (( wa4 = (doublereal *) calloc ((unsigned int) m, (unsigned int) sizeof(doublereal))) == NULL ) {
       fprintf(stderr,"calloc: Cannot allocate workspace wa4\n");
       exit(-1);
    }

    /* use the current calibration and camera constants as a starting point */
    x[0] = R->c[0];
    x[1] = R->c[1];
    x[2] = R->c[2];
    x[3] = cpar->S.c[1][0];
    x[4] = cpar->S.c[2][0];
    x[5] = cpar->S.c[3][0];
    x[6] = cpar->kappa;
    x[7] = cpar->f;
    x[8] = cpar->sx;
    x[9] = cpar->Cx;
    x[10] = cpar->Cy;

    /* define optional scale factors for the parameters */
    if ( mode == 2 ) {
        for (i = 0; i < NPARAMS; i++)
            diag[i] = 1.0;             /* some user-defined values */
    }


    auto void tf_ncc_full_optimization_error (integer *m_ptr, integer *n_ptr, doublereal *params, doublereal *err, integer *iflag);
    
    void tf_ncc_full_optimization_error (integer *m_ptr, integer *n_ptr, doublereal *params, doublereal *err, integer *iflag) {
         assert(*m_ptr == m); assert(*n_ptr == n);
         ncc_nic_optimization_error (m_ptr, n_ptr, x, fvec, cdat, cpar);
    }
    
    
    /* perform the optimization */
    lmdif (tf_ncc_full_optimization_error,  &m, &n, x, fvec, &ftol, &xtol, &gtol, &maxfev, &epsfcn,
            diag, &mode, &factor, &nprint, &info, &nfev, fjac, &ldfjac, ipvt, qtf, wa1, wa2, wa3, wa4);

    /* update the calibration and camera constants */
    R->c[0] = x[0];
    R->c[1] = x[1];
    R->c[2] = x[2];
    tf_solve_RPY_transform (cpar, R);

    cpar->S.c[1][0] = x[3];
    cpar->S.c[2][0] = x[4];
    cpar->S.c[3][0] = x[5];
    cpar->kappa = x[6];
    cpar->f = x[7];
    cpar->sx = x[8];
    cpar->Cx = x[9];
    cpar->Cy = x[10];

    /* release allocated workspace */
    free (fvec);
    free (fjac);
    free (wa4);

#ifdef DEBUG
    /* print the number of function calls during iteration */
    fprintf(stderr,"info: %d nfev: %d\n\n",info,nfev);
#endif

#undef NPARAMS
}




