#define _GNU_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <math.h>
#include <pswr.h>
#include <jsfile.h>
#include <tf_camera.h>
#include <tf_errors.h>
#include <tf_targets.h>
#include <tf_calib.h>
#include <tf_camera_plot.h>

void tf_plot_cameras
  ( PSStream *ps, 
    char *name, 
    int nmarks,
    r3_t p_w[], 
    int ncpars,
    camera_parameters_t cpar[],
    r2_t p_i[], 
    double hSize,
    double vSize,
    double hMrg,
    double vMrg,
    double Nx,
    double Ny,
    double mRadius,
    int mStyle ) 
{
  frgb_t mColor = (frgb_t){{ 1.000, 1.000, 0.400 }}; /* Mark fill color */
  pswr_new_canvas(ps, name);
  /* Negate all Y coordinates to get the proper orientation for the Y axis: */

  pswr_set_window(ps, 0.0, Nx,  -Ny, 0.0, hMrg, hSize+hMrg, vMrg, vSize+vMrg);
  /* Draw a frame around the image: */
  pswr_set_pen(ps, 0.0,0.0,0.0, 0.10, 0,0); /* Black lines */
  pswr_frame(ps);
  
  if (p_w != NULL) {
    /* Plot computed image positions: */
    r2_t *p_i_aux = (r2_t *)notnull(malloc(nmarks*sizeof(r2_t)), "no mem");
    int ip;
    for (ip = 0; ip < ncpars; ip++) {
      if (cpar[ip] != NULL) {
        int mark;
        for (mark = 0; mark < nmarks; mark++) {
          r3_t pwm = p_w[mark];
          fprintf(stderr, "p_w[%d] = ( %f %f %f )\n", mark, pwm.c[0], pwm.c[1], pwm.c[2]);
          p_i_aux[mark] = tf_world_coords_to_image_coords(cpar[ip], pwm);
          r2_t pim = p_i_aux[mark];
          fprintf(stderr, "p_i[%d] = ( %f %f )\n", mark, pim.c[0], pim.c[1]);
        }
        tf_plot_marks(ps, nmarks, p_i_aux, mRadius, mStyle + ip, mColor);
      }
    }
  }
  if (p_i != NULL) {
    /* Plot the given image coordinates: */
    tf_plot_marks(ps, nmarks, p_i, mRadius, mStyle + ncpars, mColor);
  }
}

void tf_plot_marks
  ( PSStream *ps, 
    int nmarks,
    r2_t p_i[], 
    double mRadius,
    int mStyle,
    frgb_t mColor )
  {    
    pswr_set_pen(ps, 0.0,0.0,0.0, 0.10, 0,0);
    pswr_set_fill_color(ps, mColor.c[0], mColor.c[1], mColor.c[2]);
    int mark;
    for (mark = 0; mark < nmarks; mark++) {
      r2_t pm = p_i[mark];
      switch (mStyle) {
      case 0:
        pswr_square(ps, pm.c[0], -pm.c[1], mRadius, TRUE, TRUE);
	break;
      case 1:
        pswr_dot(ps, pm.c[0], -pm.c[1], mRadius, TRUE, TRUE);
	break;
      default:
        pswr_cross(ps, pm.c[0], -pm.c[1], mRadius);
      }
    }
  }

void pswr_cross(PSStream *ps, double xc, double yc, double radius)
  {
    pswr_tic(ps, 0, xc, yc, 2*radius, 0.5);
    pswr_tic(ps, 1, xc, yc, 2*radius, 0.5);
  }
