#ifndef tf_errors_H
#define tf_errors_H

/* Functions to summarize the camera calibration errors. */

#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <math.h>
#include <ctype.h>
#include <assert.h>
#include <f2c.h>
#include <lmdif.h>
#include <r3.h> 
#include <r2.h> 
#include <r4x4.h> 
#include <rmxn.h> 
#include <affirm.h>
#include "tf_camera.h"
#include "tf_calib.h"

void tf_calib_summarize_errors (calibration_data_t cdat, camera_parameters_t cpar, FILE *ferr);
  /* This routine prints summaries of the image-space and world-space errors of the 
    camera model {cpar}, for the calibration data {cdat}. */

void tf_calib_summarize_image_errors (calibration_data_t cdat, camera_parameters_t cpar, FILE *ferr);
/* This routine prints to {ferr} the mean, standard deviation, max,
  and sum-of-squared errors in distorted image coordinates, as
  computed by {tf_camera_compute_image_error}, for the camera model
  {cpar} and calibration data {cdat}. */

void tf_calib_summarize_world_errors (calibration_data_t cdat, camera_parameters_t cpar, FILE *ferr);
/* This routine prints to file {ferr} the mean, standard deviation, max, and sum-of-squared
* error of the distance of closest approach (i.e. 3D error) between the point   *
* in object space and the line of sight formed by back projecting the measured  *
* 2D coordinates out through the camera model.                                  *
* The calculation is for all of the points in the calibration data set.         *
 */
void tf_object_space_error_stats (  calibration_data_t cdat, 
                                    camera_parameters_t cpar, 
                                    FILE *ferr  );
#endif
